module CYCSRS.CallsParam where
 
import System.Directory

confnotfound conf = do
    putStrLn $ "expected file " ++ conf ++ " not found!"
    putStrLn $ "creating default " ++ conf 
    createConf
    putStrLn $ "MANUALLY ADJUST the pathes in " ++ conf
    
createConf =
 writeFile "conf.sh" confinput
 
confinput = unlines $
 ["#!/bin/bash"
 ,"# CONFIGURATION"
 ,"# -------------"
 ,"#"
 ,"# The whole configuration is controlled by this file. "
 ,"#"
 ,"# DO NOT MODIFY OTHER FILES."
 ,"#"
 ,"# Set the following paths/values to your local installation"
 ,"#"
 ,"#"
 ,"# PATHS"
 ,"# -----"
 ,"#"
 ,"# path to tdmi"
 ,"export TDMIPATH=dist/build/TDMI"
 ,"# path to cycnt"
 ,"export CYCNTPATH=dist/build/Cycnt"
  ,"# path to the binaries of your Java installation (JDK is required, JRE is not sufficient)"
 ,""
 ,"export JAVAPATH=$PATH"
 ,""
 ,"# path to the folder that contains the minisat binary (the binary must be named minisat)"
 ,""
 ,"export MINISATPATH=$PATH"
 ,""
 ,"# path to the folder that contains the yices binary (the binary must be named yices)"
 ,""
 ,"export YICESPATH=$PATH"
 ,""
 ,"# path to the folder that contains the yices binary (the binary must be named yices-smt)"
 ,""
 ,"export YICES2PATH=$PATH"
 ,""
 ,"# path to the directory that contains the z3 binary (the binary must be named z3)"
 ,""
 ,"export Z3PATH=$PATH"
 ,""
 ,"# path to the directory that contains aprove.jar"
 ,""
 ,"export APROVEJARPATH=$PATH"
 ,""
 ,"# path to the directory that contains torpacyc (the binary must be named torpacyc)"
 ,""
 ,"export TORPACYCPATH=$PATH"
 ,""
 ,"# path to the directory that contains ttt2 (the binary must be named ttt2)"
 ,""
 ,"export TTT2PATH=$PATH"
 ,""
 ,"#"
 ,"# VALUES"
 ,"# ------ "
 ,""
 ,"# maximal memory use"
 ,""
 ,"export MEMLIMIT=6000000"
 ,""
 ,"# maximal virtual memory"
 ,""
 ,"export VMEMLIMIT=15000000"
 ,""
 ,"# maximal memory for java option Xmx"
 ,""
 ,"export JAVAMEM=6000m"
 ,""]

 
 
  
mkCallAprove maybetimeout = mkCallScript "call_aprove.sh" (callAproveScript maybetimeout)
mkCallTDMI maybetimeout = mkCallScript "call_tdmi.sh" (callTDMIScript maybetimeout)
mkCallCycnt maybetimeout = mkCallScript "call_cycnt.sh" (callCycntScript maybetimeout)

mkCallTTT2 maybetimeout = do
   mkCallScript "call_ttt2.sh" (callttt2 maybetimeout)
   case maybetimeout of
     Just x -> writeFile "comp.conf" (compconf x)
     Nothing -> writeFile "comp.conf" (compconf (300::Int))

 
mkCallTorpacyc maybetimeout = do
 mkCallScript "call_torpacyc.sh" (callTorpacycScript maybetimeout)
 writeFile "param" param2

mkCallScript scriptname script =
 do
  dir <- getCurrentDirectory
  let conf= dir ++ "/conf.sh"
  check <- doesFileExist conf
  if check then do
     writeFile scriptname (script $ conf)
     p <- getPermissions scriptname
     setPermissions scriptname (setOwnerExecutable True p)
   else do
     confnotfound conf
     error "STOPPED!"

mkCallYicesSMT2 infile outfile scriptname maybetimeout = mkCallScript (scriptname) (callYicesSMT2Script infile outfile maybetimeout)     
callYicesSMT2Script infile outfile maybetimeout pathtoconf =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in unlines 
  [ "#!/bin/bash"
  , "source " ++ pathtoconf
  , "export PATH=$JAVAPATH:$YICES2PATH:$MINISATPATH:$YICESPATH:$Z3PATH:$PATH"
  , (timeo ++ "$YICESSMT2PATH/yices-smt2 " ++ infile ++ " > " ++ outfile)
  ]
   
callCycntScript maybetimeout pathtoconf =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in unlines 
  [ "#!/bin/bash"
  , "source " ++ pathtoconf
  , "export PATH=$JAVAPATH:$YICES2PATH:$MINISATPATH:$YICESPATH:$Z3PATH:$PATH"
  , (timeo ++ "$CYCNTPATH/Cycnt $1" ++ timea)
  ] 
   
callTDMIScript maybetimeout pathtoconf =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in unlines 
  [ "#!/bin/bash"
  , "source " ++ pathtoconf
  , "export PATH=$JAVAPATH:$YICES2PATH:$MINISATPATH:$YICESPATH:$Z3PATH:$PATH"
  , (timeo ++ "$TDMIPATH/TDMI $1" ++ timea )
  ] 
callAproveScript maybetimeout pathtoconf =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in unlines 
  [ "#!/bin/bash"
  , "source " ++ pathtoconf
  , "export PATH=$JAVAPATH:$YICES2PATH:$MINISATPATH:$YICESPATH:$Z3PATH:$PATH"
  , (timeo ++ "java -Xmx6G -jar $APROVEJARPATH/aprove.jar -p plain -m wst $1" ++ timea)
  ] 
  
callTorpacycScript maybetimeout pathtoconf =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in  
  unlines
    [ "#!/bin/bash"
    , "source " ++ pathtoconf
    , "export PATH=$YICES2PATH:$PATH"
    , (timeo ++ "$TORPACYCPATH/torpacyc $1")
    ]

  

  
param2 =
 unlines
  [
  "yices-smt"
  ,"0 : 1 = verbose, 0 = not"
  ,"3 : max type graph weight (0 = no bound)"
  ,"3 : max tropical dimension"
  ,"3 : max arctic dimension"
  ,"1 : 1 = matchb, 0 = not"
  ,"3 : max mulplus dimension"
  ]

callttt2 maybetimeout pathtoconf  =
 let (timeo,timea) =  case maybetimeout of
               Nothing -> ("","")
               Just x  -> ("timeout " ++ (show x) ++ " ", " " ++ show x)
 in    
 unlines
  [ "#!/bin/bash"
  , "source " ++ pathtoconf
  , "export PATH=$YICESPATH:$MINISATPATH:$YICES2PATH:$PATH"  
  ,  (timeo ++ "$TTT2PATH/ttt2 -c comp.conf  -s COMP $1")
  ]
  
   
   
compconf tim =
  unlines $ 
   [
    "# designed for the 2011 competition"
    ,"PRETRS = ((\\"
    ," matrix -dim 1 -ib 3 -ob 5 || \\"
    ," matrix -dim 2 -ib 2 -ob 3 || \\"
    ," matrix -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -dim 3 -ib 1 -ob 1 || \\"
    ," fail)[1]*)"
    ,"DIRECTTRS = ((\\"
    ," kbo || \\"
    ," (lpo | ref;lpo) || \\"
    ," bounds -rfc -qc || \\"
    ," fail)*[7])!"
    ,"SLTRS = ((\\"
    ," sl -dp -ur -dim 1 -ib 1 -ob 2 -pl || \\"
    ," sl -dp -ur -dim 1 -ib 2 -ob 2 -pl || \\"
    ," sl -dp -ur -dim 2 -ib 1 -ob 2 -pl || \\"
    ," fail))"
    ,"ARCTICTRS = \\"
    ,"(arctic -dp -ur -dim 1 -ib 4 -ob 3[1] | \\"
    ," arctic -dp -ur -dim 2 -ib 2 -ob 2[2] | \\"
    ," arctic -dp -ur -dim 3 -ib 1 -ob 1[2] | \\"
    ," arctic -dp -ur -dim 4 -ib 1 -ob 1[2] | \\"
    ," fail)"
    ,"ARCTICBZTRS = \\"
    ,"(arctic -bz -dp -ur -dim 1 -ib 4 -ob 3[1] | \\"
    ," arctic -bz -dp -ur -dim 2 -ib 2 -ob 2[2] | \\"
    ," arctic -bz -dp -ur -dim 3 -ib 1 -ob 1[2] | \\"
    ," arctic -bz -dp -ur -dim 4 -ib 1 -ob 1[2] | \\"
    ," fail)"
    ,"BOUNDS = (bounds -dp -rfc -qc || bounds -dp -all -rfc -qc || bounds -rfc -qc)"
    ,"MAINTRS = (dp;tdg?;(sccs | sc)*;edg[0.5]?;cdg -rfc[0.5]?;((sccs | \\"
    ," (sc || sct || \\"
    ," {ur?;( \\"
    ,"  matrix -dp -ur -dim 1 -ib 3 -ob 5 || \\"
    ,"  matrix -dp -ur -dim 1 -ib 3 -ob 8 -rat 2 -db 1 || \\"
    ,"  matrix -dp -ur -dim 1 -ib 4 -ob 10 -rat 4 -db 1 || \\"
    ,"  matrix -dp -ur -dim 2 -ib 2 -ob 3 || \\"
    ,"  matrix -dp -ur -dim 3 -ib 1 -ob 3 || \\"
    ,"  matrix -dp -ur -dim 3 -ib 2 -ob 3 || \\"
    ,"  matrix -dp -ur -dim 4 -ib 1 -ob 2 || \\"
    ,"  (kbo -ur -af | lpo -ur -af) || \\"
    ,"  ARCTICTRS || \\"
    ,"  ARCTICBZTRS || \\"
    ,"  (if duplicating then fail else SLTRS) || \\"
    ,"  fail) \\"
    ," }restore || \\"
    ," BOUNDS || \\"
    ," fail \\"
    ," ))[6] | \\"
    ," uncurryx -appsym max -aarity max?;uncurryx -top -appsym max -aarity max \\"
    ," )*[" ++ show (tim `div` 6) ++ "])"
    ,"TOTALTRS=(PRETRS;(MAINTRS || DIRECTTRS || unfold))"
    ,"TRS = ((\\"
    ," if applicative then \\"
    ,"  (lpo | uncurry?;uncurry -mirror?;TOTALTRS) \\"
    ," else \\"
    ,"  TOTALTRS \\"
    ," )[" ++ show (tim `div` 6) ++ "] || \\"
    ," poly -nl -heuristic 1 -neg -ib 3 -ob 7 -direct[9] \\"
    ," )!"
    ,""
    ,"SN=(PRETRS;(MAINTRS || DIRECTTRS))"
    ,""
    ,"REV = rev?"
    ,"PRESRS = ( \\"
    ,"  matrix -dim 1 -ib 4 -ob 5[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  matrix -dim 2 -ib 2 -ob 3[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  matrix -dim 3 -ib 1 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  matrix -dim 3 -ib 2 -ob 3[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  matrix -dim 4 -ib 1 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;matrix -dim 1 -ib 4 -ob 5[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;matrix -dim 2 -ib 2 -ob 3[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;matrix -dim 3 -ib 1 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;matrix -dim 3 -ib 2 -ob 3[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;matrix -dim 4 -ib 1 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  bounds -rfc -qc[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  REV;bounds -rfc -qc[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  kbo [" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  unfold [" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  arctic -dim 1 -ib 4 -ob 5[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  arctic -dim 2 -ib 2 -ob 3[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  arctic -dim 3 -ib 1 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  arctic -dim 3 -ib 2 -ob 2[" ++ show (tim `div` 15) ++ "]|| \\"
    ,"  fail)*"
    ,"INTERNSRS = \\"
    ," matrix -dim 1 -dp -ib 5 -ob 6[" ++ show (tim-1) ++ "] || \\"
    ," matrix -dim 1 -dp -ur -ib 4 -ob 10 -rat 2 -db 5[" ++ show (tim-1) ++ "] || \\"
    ," (kbo | kbo -af -ur | lpo -af -ur)[" ++ show (tim-1) ++ "] || \\"
    ," matrix -dp -dim 2 -ib 3 -ob 3[" ++ show (tim-1) ++ "] || \\"
    ," (matrix -dp -dim 3 -ib 1 -ob 2[" ++ show (tim `div` 2) ++ "] | matrix -dp -dim 3 -ib 2 -ob 3)[" ++ show (tim-1) ++ "] || \\"
    ," matrix -dp -dim 4 -ib 1 -ob 1[" ++ show (tim-1) ++ "] || \\"
    ," matrix -dp -dim 5 -ib 1 -ob 2[" ++ show (tim-1) ++ "] ||"
    ,"ARCTICSRS = \\"
    ," arctic -dp -ur -dim 1 -ib 4 -ob 5[" ++ show (tim-1) ++ "] || \\"
    ," arctic -dp -ur -dim 2 -ib 2 -ob 3[" ++ show (tim-1) ++ "] || \\"
    ," arctic -dp -ur -dim 3 -ib 1 -ob 2[" ++ show (tim-1) ++ "] || \\"
    ," arctic -dp -ur -dim 4 -ib 1 -ob 1[" ++ show (tim-1) ++ "] || "
    ,"ARCTICBZSRS = \\"
    ," arctic -bz -dp -ur -dim 1 -ib 4 -ob 5[" ++ show (tim-1) ++ "] || \\"
    ," arctic -bz -dp -ur -dim 2 -ib 2 -ob 3[" ++ show (tim-1) ++ "] || \\"
    ," arctic -bz -dp -ur -dim 3 -ib 1 -ob 2[" ++ show (tim-1) ++ "] || \\"
    ," arctic -bz -dp -ur -dim 4 -ib 1 -ob 1[" ++ show (tim-1) ++ "] || "
    ,"BOUNDSRS = \\"
    ," ((bounds -dp -rfc -qc[" ++ show (tim `div` 3) ++ "] | bounds -dp -rfc -all -qc)[" ++ show (tim `div` 2) ++ "] || bounds -dp -qc[" ++ show (tim `div` 2) ++ "]) || "
    ,"LOOPSRS = (\\"
    ," loop -dp -r 15 -c 15[" ++ show (tim `div` 6) ++ "] | \\"
    ," loop -dp -r 20 -c 15[" ++ show (tim `div` 6) ++ "] | \\"
    ," loop -dp -r 15 -c 20[" ++ show (tim `div` 6) ++ "] | \\"
    ," loop -dp -r 20 -c 20[" ++ show (tim `div` 2) ++ "]) || "
    ,"LABSRS = (((dp;(tdg?;(sccs | ur | poly -ib 1 -ob 2 | poly -dp -ur -ib 1 -ob 3)*;rlab)2*)[3]))!"
    ,"MAINSRS = (dp;(tdg?;edg[0.5]?;cdg -rfc[0.5]?;sccs?;({ur?;( \\"
    ," INTERNSRS \\"
    ," BOUNDSRS \\"
    ," ARCTICBZSRS \\"
    ," ARCTICSRS \\"
    ," LOOPSRS \\"
    ," fail)}restore?)[" ++ show (tim-1) ++ "])*)"
    ,"SRS = ((PRESRS;(MAINSRS[" ++ show (tim-1) ++ "] || REV;MAINSRS[" ++ show (tim-1) ++ "])) || (sleep -t 20?;LABSRS))"
    ,""
    ,"#TODO: add methods in direct setting"
    ,"#DERIVATION LENGTH"
    ,"ARCTIC_CPXB = ( \\"
    ," arctic -cp -dim 1 -ib 3 -ob 4 || \\"
    ," arctic -cp -dim 1 -ib 4 -ob 5 || \\"
    ," arctic -cp -dim 2 -ib 2 -ob 3 || \\"
    ," arctic -cp -dim 2 -ib 3 -ob 4 || \\"
    ," arctic -cp -dim 3 -ib 2 -ob 3 || \\"
    ," arctic -cp -dim 3 -ib 3 -ob 4 || \\"
    ," fail)"
    ,"WGPB = ( \\"
    ," matrix -strict -cp -triangle -dim 1 -ib 4 -ob 5 || \\"
    ," matrix -strict -cp -triangle -dim 1 -ib 5 -ob 6 || \\"
    ," matrix -strict -cp -triangle -dim 1 -ib 6 -ob 7 || \\"
    ," matrix -strict -cp -triangle -dim 1 -ib 7 -ob 8 || \\"
    ," matrix -strict -cp -triangle -dim 1 -ib 8 -ob 9 || \\"
    ," fail)"
    ,"MATRIXAB = ( \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 2 -ob 3 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 4 -ob 5 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 5 -ob 5 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 4 -ib 2 -ob 2 || \\"
    ," fail)"
    ,"LINEARB = ( \\"
    ," bounds -cp[5] || \\"
    ," (if dummy then ARCTIC_CPXB else fail) || \\"
    ," MATRIXAB || \\"
    ," fail)"
    ,"QUADRATICB = ( \\"
    ," matrix -cp -overlay -triangle -direct -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 2 -ob 3 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 4 -ob 5 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 5 -ob 5 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 4 -ib 1 -ob 2 || \\"
    ," fail)"
    ,"CUBICB = ( \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 3 -ob 4 || \\"
    ," matrix -cp -mul -dim 3 -ib 3 -ob 3 || \\"
    ," fail)"
    ,"QUARTICB= ( \\"
    ," matrix -cp -overlay -triangle -dim 4 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 4 -ib 2 -ob 2 || \\"
    ," matrix -cp -mul -dim 4 -ib 2 -ob 3 || \\"
    ," fail)"
    ,"COMPLABB= rlab; ( \\"
    ," matrix -cp -overlay -triangle -direct -dim 1 -ib 2 -ob 3[15] | \\"
    ," matrix -cp -overlay -triangle -direct -dim 2 -ib 2 -ob 4[15] | \\"
    ," matrix -cp -overlay -triangle -direct -dim 3 -ib 2 -ob 2[" ++ show (tim `div` 2) ++ "] | \\"
    ," fail)"
    ,"COMPJSR= (\\"
    ," matrix -cp -jsr -dim 2 -ib 3 -ob 4[55] || \\"
    ," matrix -cp -jsr -dim 3 -ib 3 -ob 3[55] || \\"
    ," matrix -cp -jsr -dim 4 -ib 2 -ob 3[55] || \\"
    ," fail)"
    ,"COMPLEXDIRECTB = ( \\"
    ," WGPB[5]*;( \\"
    ,"    LINEARB[" ++ show (tim `div` 6) ++ "] || \\"
    ,"    (QUADRATICB[55]) || \\"
    ,"    (CUBICB[55]) || \\"
    ,"    (QUARTICB[55]) || \\"
    ,"    (COMPJSR[55]) || \\"
    ,"    fail)* \\"
    ," )"
    ,"#"
    ,"#sleep -t 5?;matrix -triangle -direct -dim 2 -ib 3 -ob 4 || \\"
    ,"#sleep -t 10?;matrix -triangle -direct -dim 3 -ib 3 -ob 3 || \\"
    ,"#sleep -t 20?;matrix -triangle -direct -dim 4 -ib 2 -ob 3 || \\"
    ,"COMPLEXITY = (var | (if duplicating then fail else ( \\"
    ," bounds [5] || bounds -qc[" ++ show (tim `div` 6) ++ "] || \\"
    ," ((if applicative then uncurry?;uncurry -mirror? else fail)?; ( \\"
    ,"  (cp;(COMPLEXDIRECTB || sleep -t 5?;COMPLABB)) || \\"
    ,"  fail))!)))"
    ,""
    ,"#certified complexity"
    ,"LINEARB_CERT = ( \\"
    ," matrix -cp -dimbound -triangle -dim 1 -ib 2 -ob 2 || \\"
    ," matrix -cp -dimbound -triangle -dim 1 -ib 4 -ob 8 || \\"
    ," matrix -cp -dimbound -triangle -dim 1 -ib 6 -ob 10 || \\"
    ," fail)"
    ,"QUADRATICB_CERT = ( \\"
    ," matrix -cp -dimbound -triangle -direct -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 2 -ob 2 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 2 -ob 3 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 3 -ob 3 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 4 -ob 5 || \\"
    ," matrix -cp -dimbound -triangle -dim 2 -ib 5 -ob 5 || \\"
    ," fail)"
    ,"CUBICB_CERT = ( \\"
    ," matrix -cp -dimbound -triangle -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -cp -dimbound -triangle -dim 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -dimbound -triangle -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -dimbound -triangle -dim 3 -ib 3 -ob 4 || \\"
    ," fail)"
    ,"QUARTICB_CERT = ( \\"
    ," matrix -cp -dimbound -triangle -dim 4 -ib 1 -ob 2 || \\"
    ," matrix -cp -dimbound -triangle -dim 4 -ib 2 -ob 2 || \\"
    ," fail)"
    ,"COMPLEX_CERT_DIRECTB = ( \\"
    ,"    ( \\"
    ,"    LINEARB_CERT[" ++ show (tim `div` 6) ++ "] || \\"
    ,"    sleep -t 3?;(QUADRATICB_CERT[55]) || \\"
    ,"    sleep -t 5?;(CUBICB_CERT[55]) || \\"
    ,"    sleep -t 5?;(QUARTICB_CERT[55]) || \\"
    ,"    fail)* \\"
    ," )"
    ,"#CeTA does currently not support \"optimization of bounds\""
    ,"COMPLEXITY_CERT = (var | (if duplicating then fail else ( \\"
    ,"  (cp;(COMPLEX_CERT_DIRECTB)) || \\"
    ,"  fail))!)"
    ,"#"
    ,"#"
    ,"#certification"
    ,"#tkbo -ib 3 -ob 4 -degree 0 -sb 2 || \\"
    ,"PRETRSCERT = ((\\"
    ," (if srs then arctic -dim 1 -ib 4 -ob 5 else fail) || \\"
    ," (if srs then arctic -dim 2 -ib 2 -ob 3 else fail) || \\"
    ," (if srs then arctic -dim 3 -ib 1 -ob 2 else fail) || \\"
    ," (if srs then arctic -dim 3 -ib 2 -ob 2 else fail) || \\"
    ," matrix -dim 1 -ib 5 -ob 8 || \\"
    ," matrix -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -dim 3 -ib 2 -ob 3 || \\"
    ," matrix -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -dim 4 -ib 1 -ob 2 || \\"
    ," matrix -dim 5 -ib 1 -ob 1 || \\"
    ," kbo -ib 3 -ob 4 || \\"
    ," fail)[5]*)"
    ,"NSNTRSCERT = ((\\"
    ," unfold || \\"
    ," fail)*[7])!"
    ,"ARCTICTRSCERT = \\"
    ,"(arctic -dp -ur -dim 1 -ib 4 -ob 3[" ++ show (tim `div` 6) ++ "] | \\"
    ," arctic -dp -ur -dim 2 -ib 2 -ob 2[15] | \\"
    ," arctic -dp -ur -dim 3 -ib 1 -ob 1[15] | \\"
    ," arctic -dp -ur -dim 4 -ib 1 -ob 1[15] | \\"
    ," fail)"
    ,"ARCTICBZTRSCERT = \\"
    ,"(arctic -bz -dp -ur -dim 1 -ib 4 -ob 3[" ++ show (tim `div` 6) ++ "] | \\"
    ," arctic -bz -dp -ur -dim 2 -ib 2 -ob 2[15] | \\"
    ," arctic -bz -dp -ur -dim 3 -ib 1 -ob 1[15] | \\"
    ," arctic -bz -dp -ur -dim 4 -ib 1 -ob 1[15] | \\"
    ," fail)"
    ,""
    ,"MAINTRSCERT = (dp;edg[1.0]?;(sccs | \\"
    ," (sc || sct || \\"
    ,"  {ur?;( \\"
    ,"   matrix -dp -ur -dim 1 -ib 3 -ob 5 || \\"
    ,"   matrix -dp -ur -dim 1 -ib 3 -ob 8 -rat 2 -db 1 || \\"
    ,"   matrix -dp -ur -dim 1 -ib 4 -ob 10 -rat 4 -db 1 || \\"
    ,"   matrix -dp -ur -dim 2 -ib 2 -ob 3 || \\"
    ,"   matrix -dp -ur -dim 2 -ib 3 -ob 4 -rat 2 -db 0 || \\"
    ,"   matrix -dp -ur -dim 3 -ib 1 -ob 3 || \\"
    ,"   matrix -dp -ur -dim 3 -ib 2 -ob 3 || \\"
    ,"   matrix -dp -ur -dim 3 -ib 2 -ob 3 -rat 2 -db 0 || \\"
    ,"   matrix -dp -ur -dim 4 -ib 1 -ob 2 || \\"
    ,"   lpo -ur -af || \\"
    ,"   ARCTICTRSCERT || \\"
    ,"   ARCTICBZTRSCERT || \\"
    ,"   fail) \\"
    ,"  } restore) \\"
    ," )*[" ++ show (tim-1) ++ "])"
    ,"SNTRSCERT=(PRETRSCERT;MAINTRSCERT)!"
    ,"TRSCERT = (SNTRSCERT || NSNTRSCERT)"
    ,""
    ,"#relative"
    ,"PRERELATIVE = (\\"
    ," (if duplicating then fail else (bounds -rt || bounds -rt -qc)) || \\"
    ," lpo -quasi || \\"
    ," matrix -dim 1 -ib 5 -ob 6 || \\"
    ," matrix -dim 2 -ib 4 -ob 5 || \\"
    ," matrix -dim 2 -ib 2 -ob 2 || \\"
    ," matrix -dim 3 -ib 3 -ob 4 || \\"
    ," matrix -dim 3 -ib 2 -ob 3 || \\"
    ," matrix -dim 4 -ib 2 -ob 3 || \\"
    ," matrix -dim 4 -ib 2 -ob 2 || \\"
    ," matrix -dim 5 -ib 2 -ob 2 || \\"
    ," matrix -dim 5 -ib 1 -ob 2 || \\"
    ," arctic -dim 1 -ib 4 -ob 5 || \\"
    ," arctic -dim 2 -ib 3 -ob 4 || \\"
    ," arctic -dim 3 -ib 2 -ob 3 || \\"
    ," arctic -dim 4 -ib 2 -ob 2 || \\"
    ," fail \\"
    ," )"
    ,"SNRELATIVE = ((\\"
    ," PRERELATIVE || \\"
    ," REV;PRERELATIVE || \\"
    ," if srs then (sleep -t 20?;rlab) else fail || \\"
    ," fail \\"
    ,")[7]*;(rt -inv -force;SN)!)[" ++ show (tim-1) ++ "]"
    ,"RELATIVE = ((\\"
    ," unfold || \\"
    ," REV;unfold || \\"
    ," SNRELATIVE))[" ++ show (tim-1) ++ "]"
    ,""
    ,"HYDRA = \\"
    ," ((emb -w;(eps0 -ib 1 -ob 2 -id 3 -od 4 -ws -d2 || eps0 -ib 1 -ob 2 -id 2 -od 3 -ws -d2)?; \\"
    ,"  (matrix -dim 2 -ib 2 -ob 2 -dp -d2 || matrix -ib 2 -ob 3 -dp -d2 -direct)*)! || \\"
    ," (dp;edg?;sccs?;(sc | eps0 -ib 1 -ob 2 -id 2 -od 3 -ws -rev; \\"
    ,"   matrix -dp -dim 1 -ib 2 -ob 2 -direct?;matrix -dp -dim 2 -ib 2 -ob 2 -direct!)))"
    ,"#everything together"
    ,"COMP = (if standard \\"
    ," then (var | con | (if srs then (SRS[" ++ show (tim-1) ++ "]) else (TRS[" ++ show (tim `div` 6) ++ "] | HYDRA))) \\"
    ," else (if relative then RELATIVE else fail))"
    ,"#complexity"
    ,"COMPLEXITY = COMPLEXITY"
    ,"ARCTIC_CPXC = ( \\"
    ," arctic -cp -dim 1 -ib 4 -ob 5 || \\"
    ," arctic -cp -dim 2 -ib 3 -ob 4 || \\"
    ," arctic -cp -dim 3 -ib 3 -ob 4 || \\"
    ," arctic -cp -dim 4 -ib 1 -ob 2 || \\"
    ," arctic -cp -dim 4 -ib 2 -ob 3 || \\"
    ," arctic -cp -dim 5 -ib 1 -ob 1 || \\"
    ," arctic -cp -dim 5 -ib 1 -ob 2 || \\"
    ," fail)"
    ,"MATRIXAC = ( \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 2 -ob 4 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 2 -ib 5 -ob 5 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 2 -ob 3 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 3 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 4 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 4 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 5 -ib 1 -ob 1 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 5 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 1 -triangle -dim 5 -ib 2 -ob 2 || \\"
    ," fail)"
    ,"LINEARC = ( \\"
    ," bounds -cp[5] || \\"
    ," (if dummy then ARCTIC_CPXC else fail) || \\"
    ," MATRIXAC || \\"
    ," fail)"
    ,"QUADRATICC = ( \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 2 -ob 4 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 3 -ob 4 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 4 -ob 5 || \\"
    ," matrix -cp -overlay -triangle -dim 2 -ib 5 -ob 5 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 3 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 3 -ib 2 -ob 3 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 4 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 4 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 5 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -bound 2 -triangle -dim 5 -ib 2 -ob 2 || \\"
    ," fail)"
    ,"CUBICC = ( \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 2 -ob 3 || \\"
    ," matrix -cp -overlay -triangle -dim 3 -ib 3 -ob 4 || \\"
    ," matrix -cp -mul -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -mul -dim 3 -ib 3 -ob 3 || \\"
    ," matrix -cp -overlay -triangle -dim 4 -bound 3 -ib 1 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 4 -bound 3 -ib 2 -ob 2 || \\"
    ," matrix -cp -overlay -triangle -dim 5 -bound 3 -ib 1 -ob 1 || \\"
    ," matrix -cp -overlay -triangle -dim 5 -bound 3 -ib 2 -ob 2 || \\"
    ," fail)"
    ,"CP1 = ((LINEARC)*) "
    ,"CP1 = (CP1[5]?;rlab;CP1)"
    ,"CP2 = ((LINEARC || QUADRATICC)*)"
    ,"CP2 = (CP2[5]?;rlab;CP2)"
    ,"CP3 = ((LINEARC || QUADRATICC || CUBICC)*)"
    ,"CP3 = (CP3[5]?;rlab;CP3)"
    ,""
    ,"#cert"
    ,"COMPCERT = ((if standard then var else fail) | con | (if srs \\"
    ," then ((sleep -t 50?;rlab;TRSCERT[" ++ show (tim-1) ++ "])! || (TRSCERT[" ++ show (tim-1) ++ "])! || (REV;SNTRSCERT[" ++ show (tim-1) ++ "])! ) \\"
    ," else (TRSCERT[" ++ show (tim-1) ++ "])))"
    ,""
    ]
    
  
