{
module CYCSRS.Parser where
import Data.Char
import Data.List
import Data.Maybe
import CYCSRS.Types
}
%name parseSequenceP Sequence
%name parseRuleP Rule
%name parseSRSP SRS
%name parseSRSRelP RelSRS
%name parseRulesP Rules
%tokentype { Token }
%token 
        cid        { TokId $$ }
        '->'       { TokSym "->"}
        '->='      { TokSym "->="}
        ','        { TokSym "," }
        '('        { TokSym "(" }
        ')'        { TokSym ")" }
        rules      { TokSym "RULES" }
%%
   
Sequence :: { Sequence }              
Sequence :  cid Sequence      {$1:$2}  
         | {- empty -} {[]}
  
Rule :: {Rule}
Rule :  Sequence '->' Sequence  {($1,$3)}
WRule :: {Rule}
WRule :  Sequence '->=' Sequence  {($1,$3)}
     
Rules :: {[Rule]}
Rules : {- empty -} {[]}
      | Rule {[$1]}
      | Rule ',' Rules {$1:$3}     
     
SRS :: {SRS}     
SRS : '(' rules Rules ')' {$3}
    | Rules {$1}

RelSRS :: {(SRS,SRS)}
RelSRS : '(' rules RelRules ')' {$3}
       | RelRules {$1}
       
RelRules :: {([Rule],[Rule])}
RelRules : {- empty -} {([],[])}
      | Rule {([],[$1])}
      | WRule {([$1],[])}
      | Rule ',' RelRules {let (a,b) = $3 in (a,$1:b) }     
      | WRule ',' RelRules {let (a,b) = $3 in ($1:a,b)}     
    
{

data Token = TokSym String | TokId String
 deriving (Eq,Show)
 
isSym (TokSym _) = True
isSym _ = False
isId x = isAlphaNum x || (x `elem` "*#$+'") 
lexId xs = let (identifier,rest) = span isId xs
           in if null identifier then Nothing else Just (identifier,rest)

startLex xs = concat $ map fill $ map lexer $ filter (not . all isSpace) (lines xs)
fill xs = if isSym (last xs) then xs else xs ++ [TokSym ","]
lexer [] = []
lexer (x:xs) 
 | x `elem` [' ', '\t'] = lexer xs
lexer ys = case (filter isJust $ map (\x -> isPrefix x ys) syms) of
          ((Just (pfx,rest)):_) -> (TokSym pfx):(lexer rest)
          [] -> case lexId ys of
                  Nothing -> error $ "unexpected symbol: " ++ (take 10 ys)  ++ "..."
                  Just (cid,r) -> (TokId cid):(lexer r)

syms = ["(",")","->=", "->",",","RULES"]                  
 

isPrefix (x:xs) (y:ys) 
 | x == y = case isPrefix xs ys of 
              Nothing -> Nothing
              Just (as,bs) -> Just (x:as,bs)
 | otherwise = Nothing      

isPrefix [] ys = Just ([],ys)
isPrefix _ [] = Nothing 

happyError :: [Token] -> a
happyError [] = error "parse error: unerwartetes Ende"
happyError xs = error ("parse error:" ++ show xs)

parseSequence = parseSequenceP . lexer
parseRule = parseRuleP . lexer
parseSRSRel = parseSRSRelP . startLex
parseSRS = parseSRSP . startLex
parseRules = parseRulesP . startLex

}
