module CYCSRS.Processor where
import CYCSRS.Types
import CYCSRS.Util.System
import Data.Maybe
import System.Clock
import Data.List

debug = False
when b c = 
  if b then c else return ()
  
type Ratio = Int

-- combine combines to processor by running them sequentially
-- the third parameter is the ratio (0 <= ratio <= 100) which
-- portions the timeout time between the first and the second processor
combine :: Processor -> Processor -> Ratio -> Processor
combine proc1 proc2 ratio =
 if ratio > 100 || ratio < 0 then error $ "in combine: ratio " ++ show ratio ++ "must be an number between 0 and 100"
  else 
  Processor {
     properties =  PProperties {cyclenontermination = 
                                  cyclenontermination (properties proc1) || cyclenontermination (properties proc2)
                                  || transformator (properties proc1)
                               ,cycletermination    = 
                                  (cycletermination (properties proc1) || cycletermination (properties proc2)) 
                                  || (transformator (properties proc1))
                               ,transformator = transformator (properties proc1) && (transformator (properties proc2))
                               ,description = description (properties proc1) ++ "--combined with--" ++ description (properties proc2)
                               },
      run = \input ->
         let tfull = timeout (options input)
             tproc1 = case tfull of 
                       Nothing -> Nothing
                       Just t -> Just $ max 1 ((ratio * t) `div` 100)
             inputproc1 = input {options = (case tproc1 of Nothing -> []; Just x -> [Timeout x]) ++ (options input)}
          in do
              -- print (description $ properties proc1)
              mtime1 <- getTime Monotonic
              output1 <- (run proc1) inputproc1
              mtime2 <- case result output1 of 
                         NO -> getTime Monotonic
                         YES -> getTime Monotonic
                         _   -> getTime Monotonic
              when debug $ putStrLn $ "DEBUG: " ++ (description $ properties proc1) ++ "TIMEOUT:" ++ (show $ timeout $ options inputproc1)
              when debug $ putStrLn $ "DEBUG: " ++ (description $ properties proc1) ++ "TIME:" ++ (show $ (timediff  mtime2 mtime1))
              let usedtime = fromIntegral $ timediff mtime2 mtime1
              let tproc2 = case tfull of 
                       Nothing -> Nothing 
                       Just t -> Just $ max (t-usedtime) (max 1 (t - (fromJust tproc1)))

              let inputproc2 = input {options = (case tproc2 of Nothing -> []; Just x -> [Timeout x]) ++ (options input)}
  --            print (description $ properties proc2)                  
              v <- case result output1 of
                    NO -> if (cyclenontermination (properties proc1)) then do
                            return output1
                           else do   
                             output2 <- ((run proc2) (inputproc2 {inputproblem = outputproblem output1, weakproblem = outputweakproblem output1}))
                             return $ output2{proof= (proof output1)++"\n"++(proof output2)}
                    YES -> if (cycletermination (properties proc1)) then do
                              return output1
                            else do 
                              output2 <- ((run proc2) (inputproc2 {inputproblem = outputproblem output1, weakproblem = outputweakproblem output1}))
                              return  $ output2{proof= (proof output1)++"\n"++(proof output2)}
                    _ -> do 
                              output2 <- ((run proc2) (inputproc2 {inputproblem = outputproblem output1, weakproblem = outputweakproblem output1}))
                              return $ output2{proof= (proof output1)++"\n"++(proof output2)}
              mtime3 <- case result v of 
                         YES -> getTime Monotonic
                         NO -> getTime Monotonic
                         _  -> getTime Monotonic
              when debug $ putStrLn $ "DEBUG: " ++ (description $ properties proc2) ++ "TIMEOUT:" ++ (show $ timeout $ options inputproc2)
              when debug $ putStrLn $ "DEBUG: " ++ (description $ properties proc2) ++ "TIME:" ++ (show $ (timediff mtime3 mtime2))
              return v
               }

timediff end start = sec $ diffTimeSpec end start           

combine_sequence :: [(Processor,Ratio)] -> Processor
combine_sequence xs = combine_it $ map (\(p,r) -> (p,fromIntegral r)) xs

combine_it :: [(Processor,Double)] -> Processor
combine_it [(p1,r1)] = p1
combine_it ((p1,r1):xs) =  combine p1 (combine_it (map (\(p,r) -> (p,100*r/(100-r1))) xs)) (round r1)


runProc relative p file timeout 
 | relative =
 do  
  (weak,strict) <- if ".srs" `isSuffixOf` file then getSRSRel file else getRelativeXML file
  let input = PInput {inputproblem = strict, weakproblem= Just weak,  options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  print rok
  print (result output)
  putStrLn (proof output)
 
 | otherwise =
 do  

  problem <- if ".srs" `isSuffixOf` file then getSRS file else getXML file
  let input = PInput {inputproblem = problem, weakproblem = Nothing,options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  print rok
  putStrLn (proof output)
 
 

runProcString relative p file timeout 
 | relative =
 do  
  (weak,strict) <- if ".srs" `isSuffixOf` file then getSRSRel file else getRelativeXML file
  let input = PInput {inputproblem = strict, weakproblem= Just weak,  options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  return (unlines [show rok,show (result output),proof output])
 
 | otherwise =
 do  

  problem <- if ".srs" `isSuffixOf` file then getSRS file else getXML file
  let input = PInput {inputproblem = problem, weakproblem = Nothing,options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  return (unlines [show rok,proof output])                

  


runProcStringNoFile relative p srs timeout 
 | relative =
 do  
  let weak = []
  let strict = srs
  let input = PInput {inputproblem = strict, weakproblem= Just weak,  options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  return (unlines [show rok,show (result output),proof output])
 
 | otherwise =
 do  
  let problem = srs
  let input = PInput {inputproblem = problem, weakproblem = Nothing,options = [Timeout timeout]}
  output <- (run p) input 
  let r = result output
  let rok = case r of
                NO  -> if (cyclenontermination (properties p)) then r else MAYBE
                YES -> if (cycletermination (properties p)) then r else MAYBE
                _ ->  r
  return (unlines [show rok,proof output])             
-- data types for termination or nontermination processors


    
data ProcessorInput  = PInput { inputproblem :: SRS, 
                                weakproblem :: Maybe SRS, -- if we use relative termination
                                options :: [Option]
                              } 
 deriving(Show)
                              
data ProcessorOutput = POutput { result  :: Result, 
                                 proof :: String, 
                                 outputproblem :: SRS,
                                 outputweakproblem :: Maybe SRS -- if we use relative termination
                               }
 deriving(Show)
 
data Result = YES | NO | MAYBE | OTHER String
instance Show Result where
 show YES = "YES"
 show NO  = "NO"
 show MAYBE = "MAYBE"
 show (OTHER s) = s
 
data Option = Timeout Int | Option String 
 deriving(Show)

data Processor = Processor 
                  { properties :: ProcessorProperties,
                    run        :: ProcessorInput -> IO ProcessorOutput
                  }

data ProcessorProperties = PProperties {cycletermination    :: Bool, 
                                        cyclenontermination :: Bool, 
                                        transformator       :: Bool,
                                        description::String}
timeout :: [Option] -> Maybe Int
timeout xs = let r = map fromTimeout $ filter isTimeout xs
             in if null r then Nothing else Just $ head r
 where isTimeout (Timeout _) = True
       isTimeout _ = False 
       fromTimeout (Timeout x) = x
