module CYCSRS.Processors.Encoders.Rotate(processor) where
import CYCSRS.Processors.Encoders.Tools
import CYCSRS.Types
import Data.List
import CYCSRS.Processor


processor = 
 Processor {
    properties = PProperties {cycletermination = False,
                              cyclenontermination=False,
                              transformator = True,
                              description="applies the transformation rotate (SZ15-RTA)"},
    run = runRotate}
 
runRotate procinput 
 | Nothing <- weakproblem procinput =
 let r =rotate2 (inputproblem procinput)
 in return $
      POutput {result=MAYBE,
               proof="PROCESSOR: ROTATE-Transformation\nApply s and complete transformation rotate (SZ15-RTA) to the problem\n"++showSRS (inputproblem procinput) ++ "\n\nOutput\n" ++ showSRS r,
               outputproblem =  r, outputweakproblem=Nothing}
 | Just wk <- weakproblem procinput =
 let (wkout,sout) =rotate2RelN (wk,(inputproblem procinput))
 in return $
      POutput {result=MAYBE,
               proof="PROCESSOR: ROTATE-Transformation\nApply sound and complete transformation rotate (SZ15-RTA) to the problem\n"++showSRSRel (wk,inputproblem procinput) ++ "\n\nOutput\n" ++ showSRSRel (wkout,sout),
               outputproblem =  sout,outputweakproblem=Just wkout}
     

  
rotate2 :: CRS -> SRS 
rotate2 crs = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
      fresh 1 = zip alph alph
      fresh 2 = freshI 'A' (map snd $ fresh 1) alph
      fresh 3 = freshI 'B' ((map snd $ fresh 1) ++ (map snd $ fresh 2)) alph
      fresh 4 = freshI 'C' (((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) alph
      fresh 5 = freshI 'D' ((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) ++ (map snd $ fresh 4)) alph
      fresh 0 = freshI 'E' (((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) ++ (map snd $ fresh 4)) ++ (map snd $ fresh 5)) (["begin","end","rewrite","cut","guess","rotate","moveleft","goright","wait","finish", "finish2"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
  in     [ ([get 0 "begin", get 0 "end"], [get 0 "rewrite",get 0 "end"])]                                               -- 1
      ++ [ ([get 0 "begin", get 1 a], [get 0 "rotate",get 0 "cut", get 4 a,get 0"guess"]) | a <- alph]                  -- 2
      ++ [ ([get 0 "guess", get 1 a], [get 4 a, get 0 "guess"]) | a <- alph]                                            -- 3
      ++ [ ([get 0 "guess", get 1 a], [get 0 "moveleft", get 3 a, get 0 "wait"]) | a <- alph]                           -- 4
      ++ [ ([get 0 "guess",get 0 "end"], [get 0 "finish",get 0 "end"])]                                                 -- 5 
      ++ [ ([get 4 b,get 0 "moveleft", get 3 a], [get 0 "moveleft",get 3 a, get 2 b]) | a <- alph, b <- alph]           -- 6
      ++ [ ([get 0 "cut", get 0 "moveleft", get 3 a], [get 5 a, get 0 "cut", get 0 "goright"]) | a <- alph]             -- 7
      ++ [ ([get 0 "goright", get 2 a], [get 4 a, get 0 "goright"]) | a <- alph]                                        -- 8
      ++ [ ([get 0 "goright", get 0 "wait", get 1 a], [get 0 "moveleft", get 3 a,get 0 "wait"])| a <- alph]             -- 9
      ++ [ ([get 0 "goright",get 0 "wait",get 0 "end"], [get 0 "finish",get 0 "end"])]                                  -- 10
      ++ [ ([get 4 a,get 0 "finish"] , [get 0 "finish", get 1 a]) | a <- alph]                                          -- 11
      ++ [ ([get 0 "cut",get 0 "finish"],[get 0 "finish2"])  ]                                                          -- 12
      ++ [ ([get 5 a,get 0 "finish2"] , [get 0 "finish2", get 1 a]) | a <- alph]                                        -- 12'
      ++ [ ([get 0 "rotate",get 0 "finish2"],[get 0 "rewrite"])]                                                       -- 13
      ++ [(get 0 "rewrite":(map (get 1) l),get 0 "begin":(map (get 1) r )) | (l,r) <- crs]                              -- 14
  
rotate2RelN :: (CRS,CRS) -> (SRS,SRS) 
rotate2RelN (weak,strict) = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) (weak ++ strict)
      fresh 1 = zip alph alph
      fresh 2 = freshI 'A' (map snd $ fresh 1) alph
      fresh 3 = freshI 'B' ((map snd $ fresh 1) ++ (map snd $ fresh 2)) alph
      fresh 4 = freshI 'C' (((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) alph
      fresh 5 = freshI 'D' ((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) ++ (map snd $ fresh 4)) alph
      fresh 0 = freshI 'E' (((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)) ++ (map snd $ fresh 4)) ++ (map snd $ fresh 5)) (["begin","end","rewrite","cut","guess","rotate","moveleft","goright","wait","finish", "finish2"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
  in (   [ ([get 0 "begin", get 0 "end"], [get 0 "rewrite",get 0 "end"])]                                               -- 1
      ++ [ ([get 0 "begin", get 1 a], [get 0 "rotate",get 0 "cut", get 4 a,get 0"guess"]) | a <- alph]                  -- 2
      ++ [ ([get 0 "guess", get 1 a], [get 4 a, get 0 "guess"]) | a <- alph]                                            -- 3
      ++ [ ([get 0 "guess", get 1 a], [get 0 "moveleft", get 3 a, get 0 "wait"]) | a <- alph]                           -- 4
      ++ [ ([get 0 "guess",get 0 "end"], [get 0 "finish",get 0 "end"])]                                                 -- 5 
      ++ [ ([get 4 b,get 0 "moveleft", get 3 a], [get 0 "moveleft",get 3 a, get 2 b]) | a <- alph, b <- alph]           -- 6
      ++ [ ([get 0 "cut", get 0 "moveleft", get 3 a], [get 5 a, get 0 "cut", get 0 "goright"]) | a <- alph]             -- 7
      ++ [ ([get 0 "goright", get 2 a], [get 4 a, get 0 "goright"]) | a <- alph]                                        -- 8
      ++ [ ([get 0 "goright", get 0 "wait", get 1 a], [get 0 "moveleft", get 3 a,get 0 "wait"])| a <- alph]             -- 9
      ++ [ ([get 0 "goright",get 0 "wait",get 0 "end"], [get 0 "finish",get 0 "end"])]                                  -- 10
      ++ [ ([get 4 a,get 0 "finish"] , [get 0 "finish", get 1 a]) | a <- alph]                                          -- 11
      ++ [ ([get 0 "cut",get 0 "finish"],[get 0 "finish2"])  ]                                                          -- 12
      ++ [ ([get 5 a,get 0 "finish2"] , [get 0 "finish2", get 1 a]) | a <- alph]                                        -- 12'
      ++ [ ([get 0 "rotate",get 0 "finish2"],[get 0 "rewrite"])]                                                       -- 13
      ++ [(get 0 "rewrite":(map (get 1) l),get 0 "begin":(map (get 1) r )) | (l,r) <- weak]                              -- 14
      ,[(get 0 "rewrite":(map (get 1) l),get 0 "begin":(map (get 1) r )) | (l,r) <- strict])
  
rotate2Rel :: (CRS,CRS) -> (SRS,SRS) 
rotate2Rel (weak,strict) = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) (weak ++ strict)
      fresh 1 = zip alph alph
      fresh 2 = freshI 'A' (map snd $ fresh 1) alph
      fresh 3 = freshI 'B' (concat [map snd (fresh i) | i<-[1,2]])  alph
      fresh 4 = freshI 'C' (concat [map snd (fresh i) | i<-[1,2,3]]) alph
      fresh 5 = freshI 'D' (concat [map snd (fresh i) | i<-[1,2,3,4]]) alph
      fresh 11 = freshI 'X' (concat [map snd (fresh i) | i<-[1,2,3,4,5]])  alph
      fresh 12 = freshI 'A' (concat [map snd (fresh i) | i<-[1,2,3,4,5,11]])  alph
      fresh 13 = freshI 'B' (concat [map snd (fresh i) | i<-[1,2,3,4,5,11,12]])  alph
      fresh 14 = freshI 'C' (concat [map snd (fresh i) | i<-[1,2,3,4,5,11,12,13]]) alph
      fresh 15 = freshI 'D' (concat [map snd (fresh i) | i<-[1,2,3,4,5,11,12,13,14]]) alph
      fresh 0 = freshI 'E' (concat [map snd (fresh i) | i<-[1,2,3,4,5,11,12,13,14,15]]) 
                            (["begin","end","rewrite","cut","guess","rotate","moveleft","goright","wait","finish", "finish2"]) 
      fresh 10 = freshI 'E' (concat [map snd (fresh i) | i<-[0,1,2,3,4,5,11,12,13,14,15]]) 
                            (["begin","end","rewrite","cut","guess","rotate","moveleft","goright","wait","finish", "finish2"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x

      sout= [ ([get 0 "begin", get 0 "end"], [get 0 "rewrite",get 0 "end"])]                                               -- 1
              ++ [ ([get 0 "begin", get 1 a], [get 0 "rotate",get 0 "cut", get 4 a,get 0"guess"]) | a <- alph]                  -- 2
              ++ [ ([get 0 "guess", get 1 a], [get 4 a, get 0 "guess"]) | a <- alph]                                            -- 3
              ++ [ ([get 0 "guess", get 1 a], [get 0 "moveleft", get 3 a, get 0 "wait"]) | a <- alph]                           -- 4
              ++ [ ([get 0 "guess",get 0 "end"], [get 0 "finish",get 0 "end"])]                                                 -- 5 
              ++ [ ([get 4 b,get 0 "moveleft", get 3 a], [get 0 "moveleft",get 3 a, get 2 b]) | a <- alph, b <- alph]           -- 6
              ++ [ ([get 0 "cut", get 0 "moveleft", get 3 a], [get 5 a, get 0 "cut", get 0 "goright"]) | a <- alph]             -- 7
              ++ [ ([get 0 "goright", get 2 a], [get 4 a, get 0 "goright"]) | a <- alph]                                        -- 8
              ++ [ ([get 0 "goright", get 0 "wait", get 1 a], [get 0 "moveleft", get 3 a,get 0 "wait"])| a <- alph]             -- 9
              ++ [ ([get 0 "goright",get 0 "wait",get 0 "end"], [get 0 "finish",get 0 "end"])]                                  -- 10
              ++ [ ([get 4 a,get 0 "finish"] , [get 0 "finish", get 1 a]) | a <- alph]                                          -- 11
              ++ [ ([get 0 "cut",get 0 "finish"],[get 0 "finish2"])  ]                                                          -- 12
              ++ [ ([get 5 a,get 0 "finish2"] , [get 0 "finish2", get 1 a]) | a <- alph]                                        -- 12'
              ++ [ ([get 0 "rotate",get 0 "finish2"],[get 0 "rewrite"])]                                                       -- 13
              ++ [(get 0 "rewrite":(map (get 1) l),get 0 "begin":(map (get 1) r )) | (l,r) <- strict]                              -- 14
      wkout = [ ([get 0 "begin", get 0 "end"], [get 10 "rewrite",get 0 "end"])]                                               -- 1
              ++ [ ([get 0 "begin", get 1 a], [get 10 "rotate",get 10 "cut", get 14 a,get 10 "guess"]) | a <- alph]                  -- 2
              ++ [ ([get 10 "guess", get 1 a], [get 14 a, get 10 "guess"]) | a <- alph]                                            -- 3
              ++ [ ([get 10 "guess", get 1 a], [get 10 "moveleft", get 13 a, get 10 "wait"]) | a <- alph]                           -- 4
              ++ [ ([get 10 "guess",get 0 "end"], [get 10 "finish",get 0 "end"])]                                                 -- 5 
              ++ [ ([get 14 b,get 10 "moveleft", get 13 a], [get 10 "moveleft",get 13 a, get 12 b]) | a <- alph, b <- alph]           -- 6
              ++ [ ([get 10 "cut", get 10 "moveleft", get 13 a], [get 15 a, get 10 "cut", get 10 "goright"]) | a <- alph]             -- 7
              ++ [ ([get 10 "goright", get 12 a], [get 14 a, get 10 "goright"]) | a <- alph]                                        -- 8
              ++ [ ([get 10 "goright", get 10 "wait", get 1 a], [get 10 "moveleft", get 13 a,get 10 "wait"])| a <- alph]             -- 9
              ++ [ ([get 10 "goright",get 10 "wait",get 0 "end"], [get 10 "finish",get 0 "end"])]                                  -- 10
              ++ [ ([get 14 a,get 10 "finish"] , [get 10 "finish", get 1 a]) | a <- alph]                                          -- 11
              ++ [ ([get 10 "cut",get 10 "finish"],[get 10 "finish2"])  ]                                                          -- 12
              ++ [ ([get 15 a,get 10 "finish2"] , [get 10 "finish2", get 1 a]) | a <- alph]                                        -- 12'
              ++ [ ([get 10 "rotate",get 10 "finish2"],[get 10 "rewrite"])]                                                       -- 13
              ++ [(get 10 "rewrite":(map (get 1) l),get 0 "begin":(map (get 1) r )) | (l,r) <- weak]                              -- 14
  in (wkout,sout)


rotate :: CRS -> SRS 
rotate crs = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
      freshsym = fresh alph (["begin","end","rewrite","cut","guess","rotate","moveleft","goright","wait","finish"])
      get [] = error "<internal error>: get called with []"
      get s = 
               case (lookup s freshsym) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
  in 
 
 [ ([get "begin", get "end"], [get "rewrite",get "end"])]
 ++
 [ ([get "begin", a], [get "rotate",get "cut", a,get "guess"]) | a <- alph]
  ++
   [([get "guess", a], [a, get "guess"]) | a <- alph]
   ++
  [([get "guess",a], [get "moveleft", a, get "wait"]) | a <- alph]
  ++
  [([get "guess",get "end"], [get "finish",get "end"])]
  ++
  [([b,get "moveleft",a], [get "moveleft",a,b]) | a <- alph, b <- alph]
  ++
  [([get "cut", get "moveleft",a], [a, get "cut", get "goright"]) | a <- alph]
  ++
  [([get "goright", a], [a, get "goright"]) | a <- alph]
  ++
  [([get "goright", get "wait", a], [get "moveleft",a,get "wait"])| a <- alph]
  ++
  [([get "goright",get "wait",get "end"], [get "finish",get "end"])
  ]
  ++
  [([a,get "finish"] , [get "finish",a]) | a <- alph]
  ++
  [([get "cut",get "finish"],[get "finish"]),([get "rotate",get "finish"],[get "rewrite"])]
  ++ [(get "rewrite":l,get "begin":r ) | (l,r) <- crs]
