module CYCSRS.Processors.Encoders.SRS2CYC(processor) where

import CYCSRS.Processor
import CYCSRS.Processors.Encoders.Tools
import CYCSRS.Types
import Data.List
 
 
processor = 
 Processor {
    properties = PProperties {cycletermination = False,
                              cyclenontermination=False,
                              transformator = False,
                              description="applies the transformation which generates an SRS from an Cycle Rewrite Relation s.t. that SRS is terminating iff the cycle rewrite relation is termination (ZBK14-RTA)"},
    run = runSRS2CYC}
 
runSRS2CYC procinput =
 return $
  POutput {result=MAYBE,
           proof="",
           outputproblem = srs2cyc (inputproblem procinput)}

 
 
srs2cyc :: SRS -> CRS 
srs2cyc crs = 
 let 
       alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
       maxnum   = maximum $ map (\(a,b) -> length a) crs
       freshsym = fresh alph (["R","E","L"] ++ [a | a <- alph])
       get [] = error "fresh with []"
       get s = case (lookup s freshsym) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
 in
  [([get "R", get "E"],[get "L", get "E"])]
  ++
  [([a, get "L"],[get "L", get a]) | a <- alph]
  ++ 
  [([get "R", get a],[a,get "R"]) | a <- alph]
  ++
  [((l ++ [get "L"]),(r ++ [get "R"])) | (l,r) <- crs]

  
