module CYCSRS.Processors.Encoders.Shift(processor) where
import CYCSRS.Processor
import CYCSRS.Processors.Encoders.Tools
import CYCSRS.Types
import Data.List


processor = 
 Processor {
    properties = PProperties {cycletermination = False,
                              cyclenontermination=False,
                              transformator=True,
                              description="applies the transformation shift (SZ15-RTA)"},
    run = runShift}
 
runShift procinput 
 | Nothing <- weakproblem procinput =
 let r = shift3 (inputproblem procinput)
 in
  return $
   POutput {result=MAYBE,
           proof="PROCESSOR: SHIFT-Transformation\nApply sound and complete transformation shift (SZ15-RTA) to the problem\n"++showSRS (inputproblem procinput) ++ "\n\nOutput\n" ++ showSRS r, outputweakproblem=Nothing,
           outputproblem =  r}
           
 | Just wk <- weakproblem procinput =
 let (weak,strict) = shift3RelN (wk,inputproblem procinput)
 in
  return $
   POutput {result=MAYBE,
           proof="PROCESSOR: SHIFT-Transformation\nApply sound and complete transformation shift (SZ15-RTA) to the problem\n"++showSRSRel (wk,inputproblem procinput) ++ "\n\nOutput\n" ++ showSRSRel (weak,strict), outputweakproblem=Just weak,
           outputproblem =  strict}

 

 
 
 
-- The encoding shift:     
    
{-
B -> W M^k V                            -- 1
M  -> eps                               -- 2
M V a -> V a'                           -- 3
a' b -> b a'                            -- 4
a' E -> a E                             -- 5
W V -> R L                              -- 6
L a -> a'' L                            -- 7
L l -> D r  for all l,r                 -- 8
a'' D -> D a                            -- 9
R D -> B                                -- 10
-}    


shift3 :: CRS -> SRS 
shift3 crs = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
      maxnum   = maximum $ map (\(a,b) -> length a) crs
      fresh 1 = zip alph alph
      fresh 2 = freshI 'X' (map snd $ fresh 1) alph
      fresh 3 = freshI 'Y' ((map snd $ fresh 1) ++ (map snd $ fresh 2)) alph
      fresh 0 = freshI 'Z' (((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)))) 
        (["B","W","M","V","R","L","D","E"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
  in     
   [([get 0 "B"], (get 0 "W":(replicate (maxnum-1) (get 0 "M"))) ++ [get 0 "V"])]
   ++
   [([get 0 "M"], [])]
   ++
   [([get 0 "M", get 0 "V", get 1 a], [get 0 "V", get 2 a]) | a <- alph]
   ++
   [([get 2 a, get 1 b], [get 1 b, get 2 a]) | a <- alph, b <- alph]
   ++
   [([get 2 a, get 0 "E"], [get 1 a, get 0 "E"]) | a <- alph]
   ++
   [([get 0 "W", get 0 "V"], [get 0 "R", get 0 "L"])]
   ++
   [([get 0 "L", get 1 a], [get 3 a, get 0 "L"]) | a <- alph]
   ++ 
   [((get 0 "L"):(map (get 1) l), (get 0 "D"):(map (get 1) r)) | (l,r) <- crs]
   ++
   [([get 3 a, get 0 "D"], [get 0 "D", get 1 a]) | a <- alph]
   ++
   [([get 0 "R", get 0 "D"],[get 0 "B"])]
 
shift3RelN (weak,strict) = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) (weak ++ strict)
      maxnum   = maximum $ map (\(a,b) -> length a) (weak ++ strict)
      fresh 1 = zip alph alph
      fresh 2 = freshI 'X' (map snd $ fresh 1) alph
      fresh 3 = freshI 'Y' ((map snd $ fresh 1) ++ (map snd $ fresh 2)) alph
      fresh 0 = freshI 'Z' (((((map snd $ fresh 1) ++ (map snd $ fresh 2)) ++ (map snd $ fresh 3)))) 
        (["B","W","M","V","R","L","D","E"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
  in     
   ([([get 0 "B"], (get 0 "W":(replicate (maxnum-1) (get 0 "M"))) ++ [get 0 "V"])]
   ++
   [([get 0 "M"], [])]
   ++
   [([get 0 "M", get 0 "V", get 1 a], [get 0 "V", get 2 a]) | a <- alph]
   ++
   [([get 2 a, get 1 b], [get 1 b, get 2 a]) | a <- alph, b <- alph]
   ++
   [([get 2 a, get 0 "E"], [get 1 a, get 0 "E"]) | a <- alph]
   ++
   [([get 0 "W", get 0 "V"], [get 0 "R", get 0 "L"])]
   ++
   [([get 0 "L", get 1 a], [get 3 a, get 0 "L"]) | a <- alph]
   ++ 
   [((get 0 "L"):(map (get 1) l), (get 0 "D"):(map (get 1) r)) | (l,r) <- weak]
   ++
   [([get 3 a, get 0 "D"], [get 0 "D", get 1 a]) | a <- alph]
   ++
   [([get 0 "R", get 0 "D"],[get 0 "B"])]
   ,
   [((get 0 "L"):(map (get 1) l), (get 0 "D"):(map (get 1) r)) | (l,r) <- strict])

shift3Rel :: (CRS,CRS) -> (SRS,SRS) 
shift3Rel (weak,strict) = 
  let alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) (weak ++ strict)
      maxnum   = maximum $ map (\(a,b) -> length a) (weak ++ strict)
      fresh 1 = zip alph alph
      fresh 2 = freshI 'X' (map snd $ fresh 1) alph
      fresh 3 = freshI 'Y' ((map snd $ fresh 1) ++ (map snd $ fresh 2)) alph
      fresh 11 = freshI 'K' ((concat [map snd $ fresh i | i <- [1,2,3]])) alph
      fresh 12 = freshI 'L' ((concat [map snd $ fresh i | i <- [1,2,3,11]])) alph
      fresh 13 = freshI 'L' ((concat [map snd $ fresh i | i <- [1,2,3,11,12]])) alph
      fresh 0 = freshI 'Z' ((concat [map snd $ fresh i | i <- [1,2,3,11,12,13]])) (["B","W","M","V","R","L","D","E"]) 
      fresh 10 = freshI 'Z' ((concat [map snd $ fresh i | i <- [1,2,3,11,12,13,0]])) (["B","W","M","V","R","L","D","E"]) 
        
      get i [] = error "<internal error>: get called with []"
      get i s = 
               case (lookup s $ fresh i) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
      strictr =
           [([get 0 "B"], (get 0 "W":(replicate (maxnum-1) (get 0 "M"))) ++ [get 0 "V"])]
           ++
           [([get 0 "M"], [])]
           ++
           [([get 0 "M", get 0 "V", get 1 a], [get 0 "V", get 2 a]) | a <- alph]
           ++
           [([get 2 a, get 1 b], [get 1 b, get 2 a]) | a <- alph, b <- alph]
           ++
           [([get 2 a, get 0 "E"], [get 1 a, get 0 "E"]) | a <- alph]
           ++
           [([get 0 "W", get 0 "V"], [get 0 "R", get 0 "L"])]
           ++
           [([get 0 "L", get 1 a], [get 3 a, get 0 "L"]) | a <- alph]
           ++
           [([get 3 a, get 0 "D"], [get 0 "D", get 1 a]) | a <- alph]
           ++
           [([get 0 "R", get 0 "D"],[get 0 "B"])]
           ++ 
           [((get 0 "L"):(map (get 1) l), (get 0 "D"):(map (get 1) r)) | (l,r) <- strict]
      weakr = 
           [([get 0 "B"], (get 10 "W":(replicate (maxnum-1) (get 10 "M"))) ++ [get 10 "V"])]
           ++
           [([get 10 "M"], [])]
           ++
           [([get 10 "M", get 10 "V", get 1 a], [get 10 "V", get 12 a]) | a <- alph]
           ++
           [([get 12 a, get 1 b], [get 1 b, get 12 a]) | a <- alph, b <- alph]
           ++
           [([get 12 a, get 0 "E"], [get 1 a, get 0 "E"]) | a <- alph]
           ++
           [([get 10 "W", get 10 "V"], [get 10 "R", get 10 "L"])]
           ++
           [([get 10 "L", get 1 a], [get 13 a, get 10 "L"]) | a <- alph]
           ++
           [([get 13 a, get 10 "D"], [get 10 "D", get 1 a]) | a <- alph]
           ++
           [([get 10 "R", get 10 "D"],[get 0 "B"])]
           ++
           [((get 10 "L"):(map (get 1) l), (get 10 "D"):(map (get 1) r)) | (l,r) <- weak]
  in (weakr,strictr)
   
   
shift :: CRS -> SRS 
shift crs = 
 let 
       alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
       maxnum   = maximum $ map (\(a,b) -> length a) crs
       freshsym = fresh alph (["B","M","E"] ++ ["R" ++  a | a <- alph])
       get [] = error "fresh with []"
       get s = case (lookup s freshsym) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
 in
  [(l, replicate (maxnum-1) (get "M") ++ r) | (l,r) <- crs]
  ++
  [([a, get "M"], [get "M", a]) | a <- alph]
  ++
  [([get "M"],[])]
  ++
  [([get "B"] ++ replicate k (get "M") ++ [a], [get "B"] ++ replicate (k-1) (get "M") ++ [ra] ) | a <- alph, k <- [1..maxnum], let ra = get ("R" ++ a)]
  ++
  [([ra, b], [b, ra]) | a <- alph, b <- alph, let ra = get ("R"++ a)]
  ++
  [([ra, get "E"], [a, get "E"]) | a <- alph,  let ra = get ("R"++ a)]  
  

  
shift2 :: CRS -> SRS 
shift2 crs = 
 let 
       alph = nub $ concat $ concatMap (\(a,b) -> [a,b]) crs
       maxnum   = maximum $ map (\(a,b) -> length a) crs
       freshsym = fresh alph (["B","M","E","W","R","L","S"])
       get [] = error "fresh with []"
       get s = case (lookup s freshsym) of
                Nothing -> (error $ "<internal error>: did not find fresh symbol for: " ++ (show s))
                Just x  -> x
 in
  [([get "L"] ++  l,  [get "R"] ++ r) | (l,r) <- crs]
  ++
  [([get "B",a], [get "W", get "L"] ++ ((replicate (maxnum-1) "M")) ++ [a]) | a <- alph]
  ++
  [([get "L", get "M"], [get "L"])]
  ++
  ([([get "M", a],[get "S",a]) | a <- alph])
  ++ 
  ([([get "S", a, b],[b, get "S",a]) | a <- alph, b <- alph])
  ++ 
  [([get "S", get "E"],[get "E"])]
  ++ 
  ([([get "L", a],[a, get "L"]) | a <- alph])
  ++ 
  ([([a,get "R"],[get "R",a]) | a <- alph])
  ++ 
  [([get "W",get "R"],[get "B"])]
   