module CYCSRS.Processors.MatrixInterpretator where
import CYCSRS.Processor
import CYCSRS.Types
import CYCSRS.Util.System
import Data.List
import System.IO
import System.Exit
import System.Environment
import System.Process
import System.Directory
import CYCSRS.CallsParam

processor = 
 Processor {
    properties = PProperties {cycletermination = True,
                              cyclenontermination=False,
                              transformator=False,
                              description="Wrapper for TDMI"},
    run = runTDMI}
  where
        runTDMI procinput 
         | Just weaksrs <- weakproblem procinput =
         do
           deleteTMPFiles
           let o = options procinput
           let maybetimeout = timeout o
           let strictsrs = inputproblem procinput
           mkCallTDMI maybetimeout
           inputfile <- getFreshSRSFile
           writeFile inputfile (showSRSRel (weaksrs,strictsrs))
           err <- getErrFile
           removeERRFile
           tmpfile <- getTempFile
           resultfile <- getOutFile
           removeOUTFile tmpfile
           let cmd = tdmi ++ " " ++ inputfile ++ " > " ++ tmpfile  ++ " 2> " ++ err
           let theproofOther  = "PROCESSOR: TDMI\nTDMI is applied to the problem\n"++showSRSRel (weaksrs,strictsrs) ++ "\n\nOutput of TDMI:\n" 
           x <- runCommand  cmd
           ecode <- waitForProcess x
           ok <- doesFileExist tmpfile
           if ok then do
             freshtmp <- getFreshFile   
             u <- runCommand $ "cp " ++ tmpfile ++ " " ++ freshtmp
             waitForProcess u
             c <- strictReadFile  freshtmp
             b <- doesFileExist freshtmp
             if b then removeFile freshtmp else return ()
             let answer = (head2 (lines c) ecode)
             let theproof  = "PROCESSOR: TDMI\nTDMI is applied to the problem\n"++showSRSRel (weaksrs,strictsrs) ++ "\n\nOutput of TDMI:\n" ++ c 
             deleteTMPFiles
             case answer of
                     YES  -> return (POutput {proof = theproof, result = answer, outputproblem = strictsrs,outputweakproblem=Just weaksrs})
                     other   -> return (POutput {proof =  theproofOther, result = answer, outputproblem = strictsrs,outputweakproblem=Just weaksrs})
            else do
                 deleteTMPFiles
                 return (POutput {proof = theproofOther, result = OTHER "Crashed", outputproblem = strictsrs,outputweakproblem=Just weaksrs})

         | Nothing <- weakproblem procinput =
         do
           deleteTMPFiles
           let o = options procinput
           let maybetimeout = timeout o
           let srs = inputproblem procinput
           mkCallTDMI maybetimeout
           inputfile <- getFreshSRSFile
           writeFile inputfile (showSRS srs)
           err <- getErrFile
           removeERRFile
           tmpfile <- getTempFile
           resultfile <- getOutFile
           removeOUTFile tmpfile
           let cmd = tdmi ++ " " ++ inputfile ++ " > " ++ tmpfile  ++ " 2> " ++ err
           let theproofOther  = "PROCESSOR: TDMI\nTDMI is applied to the problem\n"++showSRS srs ++ "\n\nOutput of TDMI:\n" 
           x <- runCommand  cmd
           ecode <- waitForProcess x
           ok <- doesFileExist tmpfile
           if ok then do
             freshtmp <- getFreshFile   
             u <- runCommand $ "cp " ++ tmpfile ++ " " ++ freshtmp
             waitForProcess u
             c <- strictReadFile  freshtmp
             b <- doesFileExist freshtmp
             if b then removeFile freshtmp else return ()
             let answer = (head2 (lines c) ecode)
             let theproof  = "PROCESSOR: TDMI\nTDMI is applied to the problem\n"++showSRS srs ++ "\n\nOutput of TDMI:\n" ++ c 
             deleteTMPFiles
             case answer of
                     YES  -> return (POutput {proof = theproof, result = answer, outputproblem = srs,outputweakproblem=Nothing})
                     other   -> return (POutput {proof =  theproofOther, result = answer, outputproblem = srs,outputweakproblem=Nothing})
            else do
                 deleteTMPFiles
                 return (POutput {proof = theproofOther, result = OTHER "Crashed", outputproblem = srs,outputweakproblem=Nothing})

        head2 [] ecode     = 
          case ecode of
             (ExitFailure 124) -> OTHER "TIMEOUT"
             _                 -> MAYBE
        head2 (x:xs) ecode = 
         let r = filter (\x -> not (x `elem` ['\r','\n']))  $ x 
         in if "Yes" `isPrefixOf` r then YES else 
             if "YES" `isPrefixOf` r then YES else
              if "NO" `isPrefixOf` r then NO else 
               if "No" `isPrefixOf` r then NO else MAYBE


        tdmi  = "./call_tdmi.sh"


        
        
        