module CYCSRS.Processors.Nontermination where
import CYCSRS.Processor
import CYCSRS.Types
import CYCSRS.Util.System
import Data.List
import System.IO
import System.Exit
import System.Environment
import System.Process
import System.Directory
import CYCSRS.CallsParam
import Control.Exception
import Data.Maybe
import qualified CYCSRS.Util.Timeout  as ST

processor = 
 Processor {
    properties = PProperties {cycletermination = False,
                              cyclenontermination=True,
                              transformator=False,
                              description="Brute-force search for cycle nontermination"},
    run = runNT}
  where
        runNT procinput 
         | Just _ <- weakproblem procinput = return 
                                            (POutput {proof =  "PROCESSOR: Nontermination-Checker\n skip - unable to prove relative termination"
                                                      ,result = MAYBE
                                                      ,outputproblem = inputproblem procinput
                                                      ,outputweakproblem = weakproblem procinput})

         | Nothing <- weakproblem procinput =
        
         do
           let o = options procinput
           let maybetimeout = timeout o
           let srs = inputproblem procinput
           let runIt = runT srs
           result <- case maybetimeout of
                       Nothing -> runIt >>= \r -> return (Just r)
                       Just t  -> ST.timeoutP (t) runIt
           case result of 
            Nothing -> return $ POutput {proof =  "PROCESSOR: Cycle-Nontermination-Checker\nCycle-Nontermination-Checker is applied to the problem\n"++showSRS srs ++ "\n\nOutput of Cycle-Nontermination-Checker:\n", result = MAYBE, outputproblem = srs, outputweakproblem=Nothing}
            Just (r,p) -> return $ POutput {proof = "PROCESSOR: Cycle-Nontermination-Checker\nCycle-Nontermination-Checker is applied to the problem\n"++showSRS srs ++ "\n\nOutput of Cycle-Nontermination-Checker:\n" ++ p, result = r, outputproblem =[],outputweakproblem=Nothing}            

runT srs =
       do
        let lmax = maximum $ map length $ map fst srs
        anyT [(from,to,i,check srs from to i) | (from,to,i) <- [(0,0,25),
                                                                (0,1,20),
                                                                (2,lmax,20),
                                                                (1*lmax+1,2*lmax,5),
                                                                (2*lmax+1,3*lmax,5),
                                                                (3*lmax+1,4*lmax,5),
                                                                (4*lmax+1,5*lmax,5),(0,10*lmax,100)]]

anyT [] = return (MAYBE,"")
anyT ((from,to,i,x):xs) =
   do (b,p) <- x
      case b of 
       NO -> return (b,p)
       _ -> anyT xs

check srs from to j = do
        let b = nonTermCheck from to j srs
        if isJust b then
          do
           return (NO, printProve srs (fromJust b))
         else
           return (MAYBE, "")


printProve srs (k,xs)  =
 do

     if k == 1 then 
        (unlines ["NO"
                , "The SRS"
                , showSRS srs
                , "is cycle-nonterminating:"
                , "found cycle rewriting loop of the form [w] o-->^+ [w]:"
                ,  ("where [w] = " ++ showS (fst $ last xs))
                , "looping rewrite sequence:"
                ]
                )
                ++
                ( let us = (reverse xs)
                      (a,b) = head us
                  in unlines $ map (\(s,i) -> ( (if i >= 0 then "\t\t o-->" ++ ( (srs)!!!i)  else "" ) ++ " " ++ showS s )) ((a,-1):(tail us)))
           else ""
           
xs !!! (-1) = ""
xs !!! i = showP $ xs !! i

showS x = "[" ++ (concat $ intersperse " " x) ++ "]"
showS1 x = (concat $ intersperse " " x)
showP (l,r) = "_{" ++ showS1 l ++ " -> " ++ showS1 r ++ "}"

nonTermCheck from to e srs 
--   | any isNonTerm (closure srs) = Just
  | otherwise = 
     let a = getAlphabet srs
         lmin = minimum $ map length $ map fst srs 
         lmax = maximum $ map length $ map fst srs
         starts = map (\x -> [(x,-1)]) $ nubBy isSim $  [lhs ++ p| lhs <- (map fst srs),  x <- [from..to], p <- perms a x]
     in   findNonTerm srs starts e 
 --     in diSteps e srs starts       
perms alpha 0   = [[]]
perms alpha len = [a:xs | a <- alpha, xs <- perms alpha (len-1)]



findNonTerm srs set depth = 
   let set' =   [n | s <- set, n <- iSteps depth srs [s]]
       res  =   [ (xs,i,k) | xs <- set', i <- [length xs-2,length xs-3..0], let (k,b) = isNonTerm (fst $ last xs,fst $ xs!!i), b] 
   in if null res then Nothing else 
    let (ys,j,k) = head res in Just $ (k,drop (j) ys)
--     anyMaybe (\xs -> isNonTerm (fst $ last xs,fst $ head xs)) set'
    
anyMaybe p [] = Nothing     
anyMaybe p (x:xs)
  | p x = Just x
  | otherwise = anyMaybe p xs


iSteps 1 srs cycles = nubBy (\ a b -> (fst $ head  a) `isSim` (fst $ head b)) [((suc,i):(s,j):xs) | ((s,j):xs) <- cycles, (suc,i) <- allOneStepCycleRewritesI srs s]   
iSteps i srs cycles = iSteps (i-1) srs [((suc,k):(s,j):xs) | ((s,j):xs) <- cycles, (suc,k) <- allOneStepCycleRewritesI srs s]

next 0 srs _ = Nothing
next i srs pathes = 
  let n = [ (suc,k) : path | path <- pathes, (suc,k) <- allOneStepCycleRewritesI srs (fst $ head path)]
      checkNT = [ m |  ((suc,k):path) <- n, let m = ntSequence (suc,k) path [], isJust m ]
  in if not (null checkNT) then (head checkNT) else next (i-1) srs n 
    

ntSequence (r,k) []           _      = Nothing
ntSequence (r,k) (x@(l,j):xs) stack
 = let (u,v) =  isNonTerm (l,r)
   in if v then --(u,((r,k):(reverse $ x:stack)))
                          Just (u,(r,k):(reverse $ x:stack))
           else ntSequence (r,k) xs (x:stack)


(^^^) :: (a -> a) -> Int ->  (a -> a)
f ^^^ 1 = f
f ^^^ i = f . (f ^^^ (i-1))
  
  
  
closure srs = iterateFix srs srs

iterateFix :: SRS -> SRS -> SRS
iterateFix srs rules =
  let rules' = rules `union`  [(l,r') | (l,r) <- rules, r' <- allOneStepCycleRewrites srs r]
  in  if rules' == rules then rules else iterateFix srs rules'

isNonTerm (l,r)
  -- | trace (show (l,r)) False  = undefined
  -- | l `isInfixOf` r   = trace (show (l,r)) True
  | r `isSim` l       = (1,True)
--   | r `isQ`   l       = (2,True)
  | otherwise         = (-1,False)
  
  
isSim w1 w2 = w1 `elem` (rotations w2)
-- wrong
isQ r l = 
 let 
     nas = sort $ nub r
     nbs = sort $ nub l
 in 
  if length nas == length nbs && nas == nbs 
   then
    if length nas == 1 
     then length r >= length l
     else if length nas == 2 
           then
            let (sym1:sym2:[]) =nas
                a_l = length [x | x <- l, x == sym1] 
                b_l = length [x | x <- l, x == sym2]
                a_r = length [x | x <- r, x == sym1]
                b_r = length [x | x <- r, x == sym2]
            in (a_l == 1 && a_r == 1 && b_r >= b_l)
               || (b_l == 1 && b_r == 1 && a_r >= a_l)
           else False 
    else False
  
-- Beispiel

test1 = printNormalForms (allOneStepCycleRewrites [(["a","a"],["a","b","a"])]) ["a","a","a"]



printRewriteSequences :: (Sequence -> [Sequence]) -> Sequence -> IO ()
printRewriteSequences rel str = putStrLn . showRewritingSequences . (allRewriteSequences rel) $ str
  where  
   showRewritingSequences [] = []
   showRewritingSequences (s:ss) = (concat . (intersperse "\n -> " )  $ (map (concat . intersperse " " ) s)) ++ "\n\n" ++ showRewritingSequences ss

printNormalForms :: (Sequence -> [Sequence]) -> Sequence -> IO ()
printNormalForms rel str =
 putStrLn $ unlines $ map (("* " ++) . showSequence) $ rewriteTrans rel str
   
allRewriteSequences :: (Sequence -> [Sequence]) -> Sequence -> [[Sequence]]
allRewriteSequences rel s = dfsP goal succ stack
 where
  goal = normalForm rel
  succ = rel 
  stack = [s]
  
rewriteTrans :: (Sequence -> [Sequence]) -> Sequence -> [Sequence]
rewriteTrans rel sequence = dfs goal succ stack
 where 
  goal = normalForm rel
  succ = rel 
  stack = [sequence]

normalForm :: (Sequence -> [Sequence]) -> Sequence -> Bool
normalForm rel sequence =  null (rel sequence)

-- =================
-- SRS
-- =================
allOneStepStringRewrites :: SRS -> Sequence -> [Sequence]
allOneStepStringRewrites srs sequence =  concat [rewriteStep rule sequence | rule <- srs]
  
-- rewriteStep takes a string rewrite rule and a string and applies
-- the rewrite rule. Since there may be several redexes all possibilities are computed.
rewriteStep :: Rule -> Sequence -> [Sequence]
rewriteStep _ [] = []
rewriteStep (l,r) sequence@(x:xs)
  | length sequence < length l = []
  | l `isPrefixOf` sequence  = 
           (r ++ (drop (length l) sequence)):[x:seq' | seq' <- rewriteStep (l,r) xs]
  | otherwise =  [x:seq' | seq' <- rewriteStep (l,r) xs]
  
-- =================
-- CycleRS  
-- =================

-- cycleAllOneStepRewrites takes a cycle rewrite system and a string computes
-- all possible successors of one cycle rewrite step
allOneStepCycleRewrites :: CRS -> Sequence -> [Sequence]
allOneStepCycleRewrites srs sequence =  concat [cycleRewriteStep rule sequence | (rule) <- srs]
  

allOneStepCycleRewritesI srs sequence =  nubBy (\(a,b) (c,d) -> isSim a c)  [ (z,i) | (rule,i) <- (zip srs [0..]), z <- cycleRewriteStep rule sequence]

-- cycleRewriteStep take a cycle rewrite rule and a string and applies the rule.
-- Since there may be several possibilities the output is a list of all possibilities
cycleRewriteStep :: Rule -> Sequence -> [Sequence]  
cycleRewriteStep (l,r) sequence =
  concat [prefixRewriteStep (l,r) s | s <- rotations sequence]
   where 
   prefixRewriteStep _ [] = []
   prefixRewriteStep (l,r) sequence@(x:xs)
    | length sequence < length l = []
    | l `isPrefixOf` sequence  = 
           (r ++ (drop (length l) sequence)):[x:seq' | seq' <- rewriteStep (l,r) xs]
    | otherwise =  []
  

-- rotations computes all rotations of a list
rotations :: [a] -> [[a]]
rotations ys = take (length ys) (iterate rot ys)
 where
        rot []     = []   
        rot (x:xs) = xs ++ [x]      
        
-- =================        
-- Depth first search

-- depth first search with implicte storing the search path 
dfsP goal succ stack = 
  go [stack]
 where
  go [[]] = []
  go [] = []
  go ([]:(a:as):xs) = go (as:xs)
  go l@((k:r):xs)
   | goal k    = (reverse $ map head l):(go (r:xs))
   | otherwise = go ((succ k):l)
  go x = error $ "in go " ++ show x
     
dfs goal succ [] = [] 
dfs goal succ (k:r) 
  | goal k     = k : (dfs goal succ r)
  | otherwise  = dfs goal succ (succ k ++ r)  

  
       
