-- A wrapper processor to call TTT2
-- No transformations are applied by this module
-- Thus the corresponding processor proves string (non)-termination

module CYCSRS.Processors.TTT2 where
import CYCSRS.Processor
import CYCSRS.Types
import CYCSRS.Util.System
import Data.List
import System.IO
import System.Exit
import System.Environment
import System.Process
import System.Directory
import CYCSRS.CallsParam


processor = 
 Processor 
   {properties = PProperties {cycletermination = False,
                              cyclenontermination=True,
                              transformator = False,
                              description="Wrapper for TTT2"},
   run = runTTT2}
 where
        runTTT2 procinput =
         do
           let o = options procinput
           let maybetimeout = timeout o
           let srs = inputproblem procinput
           let weaksrs = weakproblem procinput
           -- print weaksrs
           mkCallTTT2 maybetimeout
           inputfile <- getFreshSRSFile
           do case weaksrs of
               Nothing -> writeFile inputfile (showSRS srs)
               Just wk -> writeFile inputfile (showSRSRel (wk,srs))
           err <- getErrFile
           removeERRFile
           tmpfile <- getTempFile
           resultfile <- getOutFile
           removeOUTFile tmpfile
           let cmd = ttt2 ++ " " ++ inputfile ++ " > " ++ tmpfile  ++ " 2> " ++ err
           let theproofOther  = case weaksrs of 
                                    Nothing -> "PROCESSOR: TTT2\nTTT2 is applied to the problem\n"++showSRS srs ++ "\n\nOutput of TTT2:\n" 
                                    Just wk -> "PROCESSOR: TTT2\nTTT2 is applied to the problem\n"++showSRSRel (wk, srs) ++ "\n\nOutput of TTT2:\n"
           x <- runCommand  cmd
           ecode <- waitForProcess x
           ok <- doesFileExist tmpfile
           if ok then do
             freshtmp <- getFreshFile   
             u <- runCommand $ "cp " ++ tmpfile ++ " " ++ freshtmp
             waitForProcess u
             c <- strictReadFile  freshtmp
             b <- doesFileExist freshtmp
             if b then removeFile freshtmp else return ()
             let answer = (head2 (lines c) ecode)
             let theproof  =  case weaksrs of 
                                Nothing -> "PROCESSOR: TTT2\nTTT2 is applied to the problem\n"++showSRS srs ++ "\n\nOutput of TTT2:\n" ++ c 
                                Just wk -> "PROCESSOR: TTT2\nTTT2 is applied to the problem\n"++showSRSRel (wk, srs) ++ "\n\nOutput of TTT2:\n" ++ c 
             deleteTMPFiles
             case answer of
                     YES -> return (POutput {proof = theproof, result = answer, outputproblem = srs,outputweakproblem=weaksrs})
                     NO  -> return (POutput {proof = theproof, result = answer, outputproblem = srs,outputweakproblem=weaksrs})
                     other   -> return (POutput {proof =  theproofOther, result = answer, outputproblem = srs,outputweakproblem=weaksrs})
            else do
                 deleteTMPFiles
                 return (POutput {proof = theproofOther, result = OTHER "Crashed", outputproblem = srs,outputweakproblem=weaksrs})

        ttt2    = "./call_ttt2.sh"

        head2 [] ecode     = 
          case ecode of
             (ExitFailure 124) -> OTHER "TIMEOUT"
             _                 -> MAYBE
        head2 (x:xs) ecode = 
         let r = filter (\x -> not (x `elem` ['\r','\n']))  $ x 
         in if "Yes" `isPrefixOf` r then YES else 
             if "YES" `isPrefixOf` r then YES else
              if "NO" `isPrefixOf` r then NO else 
               if "No" `isPrefixOf` r then NO else MAYBE
