module CYCSRS.Util.System where
import System.FilePath
import System.Environment
import System.Process
import System.Directory
import System.Console.GetOpt
import Data.Time
import System.IO
import Data.IORef
import System.IO.Unsafe
import TPDB.Input
import Data.List
import TPDB.Data (lhs,rhs,rules,name,strict_rules,weak_rules)
import qualified CYCSRS.Parser as P
import CYCSRS.Types
import Control.Concurrent
import Control.Concurrent.MVar

strictReadFile file =
 do
  h <- openFile file ReadMode
  go h
  
  where 
   go h = do
           b <- hIsEOF h
           if b then hClose h >> return []
            else do
             x <- hGetChar h
             xs <- go h
             return (x:xs)
             
getSRS :: FilePath -> IO SRS
getSRS file =  do 
 con <- strictReadFile file
 let srs = (P.parseSRS con)
 eval srs
 return srs 

getSRSRel :: FilePath -> IO (SRS,SRS)
getSRSRel file =  do 
 con <- strictReadFile file
 let (weak,strict) = (P.parseSRSRel con)
 eval weak
 eval strict
 return (weak,strict)


eval :: SRS -> IO () 
eval [] = return ()
eval ((l,r):xs) =
 eval xs 
 
 
removeERRFile    =
  do
    x <- doesFileExist "err.out"
    if x then removeFile "err.out" else return ()
    system "touch err.out"
removeFileSafe file =
  do
   x <- doesFileExist file
   if x then removeFile file else return ()    
removeOUTFile file   =
  do
    x <- doesFileExist file
    if x then removeFile file  else return ()
     
-- only torpa?     
removeSRSFile = 
  do
    x <- doesFileExist "srs"
    if x then removeFile "srs" else return ()

    
    
getTempFile =
 do
  let name = "tmp_tempfile"
  p <- getPermissions "."
  if writable p then return name
                else error "cannot create temporary files in the current directory, please check permissions"

getErrFile =
 do
  let name = "err.out"
  p <- getPermissions "."
  if writable p then return name
                else error "cannot create temporary files in the current directory, please check permissions"
                
getOutFile = 
 do
  let name = "tmp_outfile"
  p <- getPermissions "."
  if writable p then return name
                else error "cannot create temporary files in the current directory, please check permissions"

getSRSFile = 
 do
  time <- getCurrentTime     
  let name = "tmp_srsfile.srs"
  p <- getPermissions "."
  if writable p then return name
                else error "cannot create temporary files in the current directory, please check permissions"
  
getLOGFile = 
 do
  time <- getCurrentTime     
  let name = "tmp_log"
  p <- getPermissions "."
  if writable p then return name
                else error "cannot create temporary files in the current directory, please check permissions"
    
{-# NOINLINE fileCount #-}

fileCount = unsafePerformIO $ newMVar 0
    
getFreshSRSFile = do
 f <- takeMVar fileCount
 putMVar fileCount (f+1)
 let name = "tmp_srsfile" ++ (show $ f+1) ++ ".srs"
 p <- getPermissions "."
 if writable p then return name
               else error "cannot create temporary files in the current directory, please check permissions"
 
getFreshFile = do
 f <- takeMVar fileCount
 putMVar fileCount (f+1)
 let name = "tmp_" ++ (show $ f+1) 
 p <- getPermissions "."
 if writable p then return name
               else error "cannot create temporary files in the current directory, please check permissions"


deleteTMPFiles =
 do
  a1 <- runCommand "touch tmp_test"
  waitForProcess a1
  a <- runCommand "rm tmp_*"
  waitForProcess a
  return ()
  sequence_ [safeRemoveFile file | file <- ["param"
                                          , "r"
                                          , "comp.conf"
                                          , "frm"
                                          , "srs"
                                          , "err.out"
                                          , "tmp_log"] ++ allScripts]

allScripts = [aprove,torpacyc,ttt2]
aprove    = "./call_aprove.sh"
torpacyc  = "./call_torpacyc.sh"
ttt2      = "./call_ttt2.sh"
-- yices     = "./call_yicesSMT2.sh"
                                          
safeRemoveFile file =
 do
    x <- doesFileExist file
    if x then removeFile file else return ()
    
getXML :: FilePath -> IO SRS
getXML file = 
 do
  x <- get_srs file
  return $ map (\x -> (map name (lhs x),map name (rhs x))) $ rules x

  
getRelativeXML :: FilePath -> IO (SRS,SRS) -- weak rules, strict rules
getRelativeXML file = 
 do
  x <- get_srs file
  let w = map (\(rhs,lhs) -> (map name rhs,map name lhs)) (weak_rules x)
  let s = map (\(rhs,lhs) -> (map name rhs,map name lhs)) (strict_rules x)
  return (w,s)
  
breakAll :: (a -> Bool) -> [a] -> [([a],[a])]
breakAll p s = 
 case break p s of 
  (r,[]) -> [(r,[])]
  q@(r,success:rest) -> q:[(r++(success:a),b) | (a,b) <- breakAll p rest]
