module TDMI.Domain.Arctic where          
import Data.Maybe
import Data.List
import Debug.Trace
import TDMI.Generic
import qualified CYCSRS.Types as T

 
arctic = Domain {
  conditionSymbol = \dim symbolmatrix  -> AND $
   ((symbolmatrix!!!(1,1)) :>: (MinInfty)):
   [(symbolmatrix!!!(row,col)) :>=: (MinInfty) | row <- [1..dim], col <- [1..dim], (row,col) /= (1,1)] 
   
 ,domainName = "arctic"   
 ,zero = MinInfty  -- MinInfty is the one of arctic domain
 ,one  = Number 0  -- 0 is the zero element of arctic domain
 ,domainToInteger = \x -> case x of 
                           MinInfty -> 1
                           (Number i) -> 2*i
 ,intPlus = \z1 z2 ->
            let 
                z1' = z1 `div` 2
                z2' = z2 `div` 2
                flag1 = z1 `mod` 2
                flag2 = z2 `mod` 2
            in if flag1 == 1 then z2 else 
                    if flag2 == 1 then z1 else
                    if z1' >= z2' then z1 else z2
 ,intMult = \z1 z2 ->
        let 
            z1' = z1 `div` 2
            z2' = z2 `div` 2
            flag1 = z1 `mod` 2
            flag2 = z2 `mod` 2
        in if flag1 == 1 then z1 else 
                if flag2 == 1 then z2 else
                2*(z1' + z2')
               
 ,lessthan = lessthanArctic
 ,equals = equalsArctic
 ,showNum =  \n -> if odd n then "-" else show (n `div` 2)
 ,fixBitWidth = fixBitWidthArctic
 ,leq = leqArctic
 ,strictsmaller = \ m1 m2 dim -> AND $
     [OR  [ (x :>: y), AND[x :=: MinInfty , y :=: MinInfty] ] | row <- [1..dim], col <- [1..dim], let x = (m1!!!(row,col)), let y=  (m2!!!(row,col))]

 ,strictlySmaller = \a b  ->
     let dim = length a
     in  all (== True) [ (lessthanArctic x y) || (x `equalsArctic` 1 && y `equalsArctic` 1) | row <- [1..dim], col <- [1..dim], let x =(a!!!(row,col)), let y = (b!!!(row,col)) ]
 ,weaklySmaller = \a b ->
     let dim = length a
     in  all (== True) [leqArctic (a!!!(row,col))  (b!!!(row,col)) | row <- [1..dim], col <- [1..dim]]
}

leqArctic = \z1 z2 -> z1 `equalsArctic` z2 || z1 `lessthanArctic` z2
lessthanArctic = \z1 z2 ->
    let 
        z1' = z1 `div` 2
        z2' = z2 `div` 2
        flag1 = z1 `mod` 2
        flag2 = z2 `mod` 2
    in if flag1 == 1 && flag2 == 0 then True
                                   else if flag1 == 0 && flag2 == 0 then z1' < z2' else False
                                   
equalsArctic = \z1 z2 ->
    let 
        z1' = z1 `div` 2
        z2' = z2 `div` 2
        flag1 = z1 `mod` 2
        flag2 = z2 `mod` 2
    in if flag1 == 1 && flag2 == 1 then True
                                   else if flag1 == flag2 then z1' == z2' else False
 
fixBitWidthArctic bitwidth (conditions,alphabet) bound = 
                    unlines $ [header] ++  declarations  ++ [conds,bounds,foot,mods]
    where
        -- header: use QF_BV logiv
        header = "(set-logic QF_BV)"

        -- make the declaration of the variables
        declarations = map (mkDeclaration (bitwidth+1)) alphabet
        mkDeclaration bwidth sym = "(declare-fun " ++ sym ++ " () (_ BitVec " ++ show bwidth ++ "))"   
        -- translate the conditions into SMT-LIB2 format
        conds = "(assert " ++ translateC bitwidth conditions ++ ")"
        -- generate conditions s.t. numbers are small enough
        bounds = "(assert " ++ translateC bitwidth boundcond ++ ")"
        boundcond = AND (map (\a -> Number (2^(bound+1)) :>: Var a) alphabet)
        -- translate the symbolic representation of conditions and arithmetic expressions into
        -- the SMT-LIB2 format
        -- shift number on pos to the left
        translateA bwidth (Number i) = "(bvmul (_ bv" ++ show i ++ " " ++ show (bwidth+1) ++ ") (_ bv2 " ++ show (bwidth+1) ++ "))"
        translateA bwidth (MinInfty) = "(_ bv1 " ++ show (bwidth+1) ++ ")"
        translateA bwidth (Var v) = fromVar (Var v)
        translateA bwidth (u :*: v)  = -- "(bvmul " ++ (translateA bwidth u) ++ " " ++ (translateA bwidth v) ++ ")"
            let u' = translateA bwidth u
                v' = translateA bwidth v
                lastbitu' = "(bvurem up (_ bv2 " ++ show (bwidth+1) ++ "))"
                lastbitv' = "(bvurem vp (_ bv2 " ++ show (bwidth+1) ++ "))"
            in
            "(let ((up " ++ u' ++ " )" ++
                "(vp " ++ v' ++ " )" ++
                ")" ++
                
                "(let ((lu " ++ lastbitu' ++ " )" ++
                        "(lv " ++ lastbitv' ++ " )" ++
                        ")" ++
                        "(ite (= lu (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++ 
                            "up " ++ 
                            "(ite (= lv (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++ 
                                "vp " ++
                                "(bvadd up vp)))" ++
                        
                        "))"
                
        translateA bwidth (u :+: v)  = 
            let u' = translateA bwidth u
                v' = translateA bwidth v
                lastbitu' = "(bvurem up (_ bv2 " ++ show (bwidth+1) ++ "))"
                lastbitv' = "(bvurem vp (_ bv2 " ++ show (bwidth+1) ++ "))"
            in
            "(let ((up " ++ u' ++ " )" ++
                "(vp " ++ v' ++ " )" ++
                ")" ++
                
                "(let ((lu " ++ lastbitu' ++ " )" ++
                        "(lv " ++ lastbitv' ++ " )" ++
                        ")" ++
                        "(ite (= lu (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++ 
                            "vp " ++ 
                            "(ite (= lv (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++ 
                                "up " ++
                                "(ite (bvult up vp) vp up)))))"
                        
        translateC bwidth (c :>: d)  = 
            let c' = translateA bwidth c
                d' = translateA bwidth d
                lastbitc' = "(bvurem cp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
                lastbitd' = "(bvurem dp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
            in    
                "(let ((cp " ++ c' ++ " )" ++
                    "(dp " ++ d' ++ " )" ++
                    ")" ++
                    
                    "(let ((lc " ++ lastbitc' ++ " )" ++
                            "(ld " ++ lastbitd' ++ " )" ++
                            ")" ++
                                "(or (bvult lc ld) " ++  -- d = -MinInfty c /= MinInfty
                                    "(and (= ld (_ bv0 " ++ show (bwidth+1) ++ "))" ++ --  d /= MinInfty
                                        "(= lc (_ bv0 " ++ show (bwidth+1) ++ "))" ++ --  c /= MinInfty
                                        "(bvult dp cp)))))\n" --                 d < c  

        translateC bwidth (c :=: d)  = 
            let c' = translateA bwidth c
                d' = translateA bwidth d
                lastbitc' = "(bvurem cp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
                lastbitd' = "(bvurem dp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
            in    
                "(let ((cp " ++ c' ++ " )" ++
                    "(dp " ++ d' ++ " )" ++
                    ")" ++
                    
                    "(let ((lc " ++ lastbitc' ++ " )" ++
                            "(ld " ++ lastbitd' ++ " )" ++
                            ")" ++
                            "(or (and (= ld (_ bv1 " ++ show (bwidth+1) ++ "))\n" ++
                                    "(= ld lc ))\n " ++  -- d = MinInfty and c = MinInfty
                                "(= cp dp))))\n" --       d = c  

        translateC bwidth (c :>=: d) = "(or " ++ translateC bwidth (c :>: d) ++ " " ++ translateC bwidth (c :=: d) ++ ")"
        translateC bwidth (NOT c)    = "(not " ++ translateC bwidth c ++ ")"
        translateC bwidth (AND cs)   = "(and " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
        translateC bwidth (OR cs)    = "(or " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
        translateC bwidth (EQUIV a b) = "(= " ++ (concat $ intersperse " " (map (translateC bwidth) [a,b])) ++ ")"
        --   maxbv b1 b2 = 
        --       "(ite (bvult " ++ b1 ++ " " ++ b2 ++ ") " ++ b2 ++ " " ++ b1 ++ ")" 
        -- check for satisfiability
        foot = "(check-sat)"
        -- print the model
        mods = "(get-value (" ++ (concat $ intersperse  " " alphabet) ++ "))"
