module TDMI.Domain.Natural where          
import Data.Maybe
import Data.List
import TDMI.Generic
import qualified CYCSRS.Types as T
    
natural = Domain {
  -- conditions for every matrix: the matrix must be of M_D, i.e.
-- the upper-left entry is > 0 and all other entries are >= 0
  conditionSymbol = \dim symbolmatrix  -> AND $
   ((symbolmatrix!!!(1,1)) :>: (Number 0)):
   [(symbolmatrix!!!(row,col)) :>=: (Number 0) | row <- [1..dim], col <- [1..dim], (row,col) /= (1,1)] 

 ,domainName = "natural"
 ,zero = Number 0
 ,one  = Number 1
 ,domainToInteger = \(Number i) -> i
 ,intMult = (*)
 ,intPlus = (+)
 ,lessthan = (<)
 ,equals  = (==)
 ,showNum = show
 ,leq = (<=)
 ,fixBitWidth = fixBitWidthNatural
 ,strictsmaller = \m1 m2 dim -> AND $
     ((m1!!!(1,1)):>: (m2!!!(1,1))):
     [(m1!!!(row,col)) :>=: (m2!!!(row,col)) | row <- [1..dim], col <- [1..dim], (row,col) /= (1,1)] 
 ,strictlySmaller = \a b -> weaklySmallerNatural a b && ((a!!!(1,1)) < (b!!!(1,1)))
 ,weaklySmaller = weaklySmallerNatural
 }
weaklySmallerNatural a b 
 = let dim = length a
   in  all (== True) [(a!!!(row,col)) <= (b!!!(row,col)) | row <- [1..dim], col <- [1..dim]]

fixBitWidthNatural bitwidth (conditions,alphabet)  bound = 
   unlines $ [header] ++  declarations  ++ [conds,bounds,foot,mods]
 where
  -- header: use QF_BV logiv
  header = "(set-logic QF_BV)"

  -- make the declaration of the variables
  declarations = map (mkDeclaration bitwidth) alphabet 
  mkDeclaration bwidth sym = "(declare-fun " ++ sym ++ " () (_ BitVec " ++ show bwidth ++ "))"   
  -- translate the conditions into SMT-LIB2 format
  conds = "(assert " ++ translateC bitwidth conditions ++ ")"
  -- generate conditions s.t. numbers are small enough
  bounds = "(assert " ++ translateC bitwidth boundcond ++ ")"
  boundcond = AND (map (\a -> Number (2^bound) :>: Var a) alphabet)
  -- translate the symbolic representation of conditions and arithmetic expressions into
  -- the SMT-LIB2 format
  translateA bwidth (Number i) = "(_ bv" ++ show i ++ " " ++ show bwidth ++ ")"
  translateA bwidth (Var v) = fromVar (Var v)
  translateA bwidth (u :*: v)  = "(bvmul " ++ (translateA bwidth u) ++ " " ++ (translateA bwidth v) ++ ")"
  translateA bwidth (u :+: v)  = "(bvadd " ++ (translateA bwidth u) ++ " " ++ (translateA bwidth v) ++ ")"
  translateC bwidth (c :>: d)  = "(bvult " ++ (translateA bwidth d) ++ " " ++ (translateA bwidth c) ++ ")"
  translateC bwidth (c :=: d)  = "(= " ++ (translateA bwidth c) ++ " " ++ (translateA bwidth d) ++ ")"
  translateC bwidth (c :>=: d) = "(or " ++ translateC bwidth (c :>: d) ++ " " ++ translateC bwidth (c :=: d) ++ ")"
  translateC bwidth (NOT c)    = "(not " ++ translateC bwidth c ++ ")"
  translateC bwidth (AND cs)   = "(and " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
  translateC bwidth (OR cs)    = "(or " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
  translateC bwidth (EQUIV a b) = "(= " ++ (concat $ intersperse " " (map (translateC bwidth) [a,b])) ++ ")"

  
  -- check for satisfiability
  foot = "(check-sat)"
  -- print the model
  mods = "(get-value (" ++ (concat $ intersperse  " " alphabet) ++ "))"


