module TDMI.Domain.Tropical where          
import Data.Maybe
import Data.List
import Debug.Trace
import TDMI.Generic
import qualified CYCSRS.Types as T

 
tropical = Domain {
  conditionSymbol = \dim symbolmatrix  -> AND $
   ((NOT ((MaxInfty) :=: (symbolmatrix!!!(1,1))))):
   [(symbolmatrix!!!(row,col)) :>=: (Number 0) | row <- [1..dim], col <- [1..dim]] 
  ,
  domainName = "tropical"   
 ,zero = MaxInfty  -- MaxInfty is the zero of tropical domain
 ,one  = Number 0  -- 0 is the one element of tropical domain
 ,domainToInteger = \x -> case x of 
                           MaxInfty -> 1
                           (Number i) -> 2*i
 ,intPlus = \z1 z2 -> -- minimum
            let 
                z1' = z1 `div` 2
                z2' = z2 `div` 2
                flag1 = z1 `mod` 2
                flag2 = z2 `mod` 2
            in if flag1 == 1 then z2 else 
                    if flag2 == 1 then z1 else
                    if z1' <= z2' then z1 else z2
 ,intMult = \z1 z2 -> -- plus
        let 
            z1' = z1 `div` 2
            z2' = z2 `div` 2
            flag1 = z1 `mod` 2
            flag2 = z2 `mod` 2
        in if flag1 == 1 then z1 else 
                if flag2 == 1 then z2 else
                2*(z1' + z2')
               
 ,lessthan = lessthanTropical
 ,equals = equalsTropical
 ,showNum =  \n -> if odd n then "+" else show (n `div` 2)
 ,fixBitWidth = fixBitWidthTropical
 ,leq = leqTropical
 ,strictsmaller = \ m1 m2 dim -> AND $
     [OR  [ (x :>: y), AND[x :=: MaxInfty , y :=: MaxInfty] ]
        | row <- [1..dim], col <- [1..dim], let x = (m1!!!(row,col)), let y = (m2!!!(row,col))] 

 ,strictlySmaller = \a b  ->
     let dim = length a
     in  all (== True) [lessthanTropical x y || (x `equalsTropical` 1 && y `equalsTropical` 1)  | row <- [1..dim], col <- [1..dim], let x = (a!!!(row,col)), let y =  (b!!!(row,col))]
 ,weaklySmaller = \a b ->
     let dim = length a
     in  all (== True) [leqTropical x y  | row <- [1..dim], col <- [1..dim], let x= (a!!!(row,col)), let y=  (b!!!(row,col))]
}

leqTropical = \z1 z2 -> z1 `equalsTropical` z2 || z1 `lessthanTropical` z2
lessthanTropical = \z1 z2 ->
    let 
        z1' = z1 `div` 2
        z2' = z2 `div` 2
        flag1 = z1 `mod` 2
        flag2 = z2 `mod` 2
    in if flag1 == 0 && flag2 == 1 then True
                                   else if flag1 == 0 && flag2 == 0 then z1' < z2' else False
                                   
equalsTropical = \z1 z2 ->
    let 
        z1' = z1 `div` 2
        z2' = z2 `div` 2
        flag1 = z1 `mod` 2
        flag2 = z2 `mod` 2
    in if flag1 == 1 && flag2 == 1 then True
                                   else if flag1 == flag2 then z1' == z2' else False
 
fixBitWidthTropical bitwidth (conditions,alphabet) bound = 
                    unlines $ [header] ++  declarations  ++ [conds,bounds,foot,mods]
    where
        -- header: use QF_BV logiv
        header = "(set-logic QF_BV)"

        -- make the declaration of the variables
        declarations = map (mkDeclaration (bitwidth+1)) alphabet
        mkDeclaration bwidth sym = "(declare-fun " ++ sym ++ " () (_ BitVec " ++ show bwidth ++ "))"   
        -- translate the conditions into SMT-LIB2 format
        conds = "(assert " ++ translateC bitwidth conditions ++ ")"
        -- generate conditions s.t. numbers are small enough
        bounds = "(assert " ++ translateC bitwidth boundcond ++ ")"
        boundcond = AND (map (\a -> Number (2^(bound+1)) :>: Var a) alphabet)
        -- translate the symbolic representation of conditions and arithmetic expressions into
        -- the SMT-LIB2 format
        -- shift number on pos to the left
        translateA bwidth (Number i) = "(bvmul (_ bv" ++ show i ++ " " ++ show (bwidth+1) ++ ") (_ bv2 " ++ show (bwidth+1) ++ "))"
        translateA bwidth (MaxInfty) = "(_ bv1 " ++ show (bwidth+1) ++ ")"
        translateA bwidth (Var v) = fromVar (Var v)
        translateA bwidth (u :*: v)  = 
         -- + 
            let u' = translateA bwidth u
                v' = translateA bwidth v
                lastbitu' = "(bvurem up (_ bv2 " ++ show (bwidth+1) ++ "))"
                lastbitv' = "(bvurem vp (_ bv2 " ++ show (bwidth+1) ++ "))"
            in
            "(let ((up " ++ u' ++ " )" ++
                "(vp " ++ v' ++ " )" ++
                ")" ++
                
                "(let ((lu " ++ lastbitu' ++ " )" ++
                        "(lv " ++ lastbitv' ++ " )" ++
                        ")" ++
                        "(ite (= lu (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++       -- if u is MaxInfty 
                            "up " ++                                                -- then u
                            "(ite (= lv (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++   -- else if v is MaxInfty
                                "vp " ++                                            --       then v
                                "(bvadd up vp)))" ++                                -- else u+v
                        
                        "))"
                
        translateA bwidth (u :+: v)  = 
            let u' = translateA bwidth u
                v' = translateA bwidth v
                lastbitu' = "(bvurem up (_ bv2 " ++ show (bwidth+1) ++ "))"
                lastbitv' = "(bvurem vp (_ bv2 " ++ show (bwidth+1) ++ "))"
            in
            "(let ((up " ++ u' ++ " )" ++
                "(vp " ++ v' ++ " )" ++
                ")" ++
                
                "(let ((lu " ++ lastbitu' ++ " )" ++
                        "(lv " ++ lastbitv' ++ " )" ++
                        ")" ++
                        "(ite (= lu (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++           -- if u is MaxInfty
                            "vp " ++                                                    -- then v
                            "(ite (= lv (_ bv1 " ++ show (bwidth+1) ++ ")) \n" ++       -- if v is MaxInfty
                                "up " ++                                                -- then v
                                "(ite (bvult up vp) up vp)))))"                         -- else if u < v then u else v
                        
        translateC bwidth (c :>: d)  = 
            let c' = translateA bwidth c
                d' = translateA bwidth d
                lastbitc' = "(bvurem cp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
                lastbitd' = "(bvurem dp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
            in    
                "(let ((cp " ++ c' ++ " )" ++
                    "(dp " ++ d' ++ " )" ++
                    ")" ++
                    
                    "(let ((lc " ++ lastbitc' ++ " )" ++
                            "(ld " ++ lastbitd' ++ " )" ++
                            ")" ++
                                "(or (bvult ld lc) " ++                                -- c = MaxInfty d /= MaxInfty
                                    "(and (= ld (_ bv0 " ++ show (bwidth+1) ++ "))" ++ --  d /= MaxInfty
                                        "(= lc (_ bv0 " ++ show (bwidth+1) ++ "))" ++ --  c /= MaxInfty
                                        "(bvult dp cp)))))\n" --                 d < c  

        translateC bwidth (c :=: d)  = 
            let c' = translateA bwidth c
                d' = translateA bwidth d
                lastbitc' = "(bvurem cp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
                lastbitd' = "(bvurem dp (_ bv2 " ++ show (bwidth+1) ++ "))\n"
            in    
                "(let ((cp " ++ c' ++ " )" ++
                    "(dp " ++ d' ++ " )" ++
                    ")" ++
                    
                    "(let ((lc " ++ lastbitc' ++ " )" ++
                            "(ld " ++ lastbitd' ++ " )" ++
                            ")" ++
                            "(or (and (= ld (_ bv1 " ++ show (bwidth+1) ++ "))\n" ++
                                    "(= ld lc ))\n " ++  -- d = MaxInfty and c = MaxInfty
                                "(= cp dp))))\n" --       d = c  

        translateC bwidth (c :>=: d) = "(or " ++ translateC bwidth (c :>: d) ++ " " ++ translateC bwidth (c :=: d) ++ ")"
        translateC bwidth (NOT c)    = "(not " ++ translateC bwidth c ++ ")"
        translateC bwidth (AND cs)   = "(and " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
        translateC bwidth (OR cs)    = "(or " ++ (concat $ intersperse " " (map (translateC bwidth) cs)) ++ ")"
        translateC bwidth (EQUIV a b) = "(= " ++ (concat $ intersperse " " (map (translateC bwidth) [a,b])) ++ ")"
        --   maxbv b1 b2 = 
        --       "(ite (bvult " ++ b1 ++ " " ++ b2 ++ ") " ++ b2 ++ " " ++ b1 ++ ")" 
        -- check for satisfiability
        foot = "(check-sat)"
        -- print the model
        mods = "(get-value (" ++ (concat $ intersperse  " " alphabet) ++ "))"
