module Main where
import System.Directory
import Data.List 
import Data.Maybe
import qualified CYCSRS.Types as T
import CYCSRS.Util.System
import CYCSRS.CallsParam
import CYCSRS.Processor
import CYCSRS.Util.System(removeFileSafe)
import System.Environment
import System.Process
import qualified System.Timeout as Timeout
import Control.Concurrent
import System.Clock
import TDMI.Generic
import TDMI.Domain.Arctic
import TDMI.Domain.Tropical
import TDMI.Domain.Natural

main =
 do 
  (a:b:cs) <- getArgs 
  (w,s) <- if "xml" `isSuffixOf` a then
            getRelativeXML a
          else getSRSRel a
  let stamp = if null cs then Nothing else Just (head cs)
  if null w then do
     (r,p,x) <- init_and_run stamp False ([],s) (Just $ read b)   
     putStrLn p
   else do
    (r,p,x) <- init_and_run stamp True (w,s) (Just $ read b)   
    putStrLn p

init_and_run :: Maybe String -> Bool -> (T.SRS,T.SRS) -> Maybe Int -> IO (Result,String,(T.SRS,T.SRS))  
init_and_run stamp relative (weaksrs,strictsrs) maybetimeout = do
  
  timestamp <- getTime Monotonic
  remainingproblem <- newMVar (weaksrs,strictsrs)
  proof <- newMVar ""
  let prefixStamp = case stamp of 
                       Nothing -> "tmp_tdmi" ++ (show $ nsec timestamp) ++ "_"
                       Just st -> st
  nameMVar <- newMVar ([prefixStamp ++ show i | i <- [1..]],[])
  let dim_bsize_list = [(1,x) | x <- [2,8]] ++ [(2,2),(2,3),(4,1),(3,2),(2,4),(5,1),(3,3),(4,2),(3,4),(4,3),(4,4)] ++ [(x,y) | x <- [2..10], y <- [1..100]] 
  let provermain maybetimeout = iterRun nameMVar relative maybetimeout proof remainingproblem (repeat  [arctic] {- natural,arctic,tropical]-} ) dim_bsize_list dim_bsize_list (weaksrs,strictsrs)
  appendStream proof "============================="
  appendStream proof "Matrix-Interpretator applied to:\n"
  appendStream proof $ (if relative then T.showSRSRel (weaksrs,strictsrs) else T.showSRS strictsrs)
  res <- case maybetimeout of 
          Nothing -> do r <- provermain Nothing
                        return $ Just r
          Just time -> do Timeout.timeout (time*(10^6)) (provermain maybetimeout)
  theproof <- takeMVar proof
  case res of
    Just result -> do
                    theproblem <- tryTakeMVar remainingproblem
                    case theproblem of 
                        Nothing -> return (result,theproof, (weaksrs,strictsrs))
                        Just tp -> return (result,theproof,tp)
    Nothing     ->  do
                    theproblem <- tryTakeMVar remainingproblem
                    case theproblem of 
                         Nothing -> return (MAYBE,theproof, (weaksrs,strictsrs))
                         Just tp -> return (MAYBE,theproof, tp)
takeMVarWhile p mvar =
    do
      c <- takeMVar mvar
      if p c  then takeMVarWhile p mvar else return c
iteratDom result nameMVar relative maybetimeout d [] (weaksrs,strictsrs) =
    putMVar result Nothing
iteratDom result  nameMVar relative maybetimeout d ((dim,bsize):dim_bsize_list) (weaksrs,strictsrs) =
  do 
    res <- runP nameMVar relative maybetimeout d dim bsize (weaksrs,strictsrs)
    case res of
     Just x -> putMVar result res 
     Nothing -> iteratDom  result nameMVar relative maybetimeout d dim_bsize_list (weaksrs,strictsrs)

-- ------------------------------------------------------
-- iterRun, iteratively tries to find matrix interpretation, removes rules and restarts                     
 
iterRun :: MVar ([String],[String]) -> Bool -> Maybe Int -> MVar String -> MVar (T.SRS,T.SRS) -> [[Domain]] -> [(Int,Int)] -> [(Int,Int)] -> (T.SRS,T.SRS) -> IO Result  
iterRun nameMVar relative maybetimeout proof _ _ _ _ ([],[]) = do
 appendStream proof "The problem is empty, thus CycleTermination has been proved. QED"
 prependStream proof "YES"
 return YES
iterRun nameMVar True maybetimeout proof _ _ _ _ (_,[]) = do
 appendStream proof "There are no strict rules, thus relative CycleTermination has been proved. QED"
 prependStream proof "YES"
 return YES

 
              

iterRun nameMVar relative maybetimeout proof mvarprob ((dom):doms) rec@(dim_bsize_list) origl (weaksrs,strictsrs) = do  
  start <- getTime Monotonic
--  appendStream proof $ "Dimension: " ++ show dim
--  appendStream proof $ "BitWidth: " ++ show bsize
  res <- 
   do 
    result <- newEmptyMVar
    threads <- mapM forkIO
            [ iteratDom  result nameMVar relative maybetimeout d rec (weaksrs,strictsrs)
                | d <- dom
            ]                
    theresult <- takeMVarWhile isNothing result
    -- print theresult
    mapM_ killThread threads
    (_,used) <- readMVar nameMVar
    sequence_ [removeFileSafe file | file <- used]
    return theresult
  end <- getTime Monotonic
  let time = sec $ diffTimeSpec end start        
  appendStream proof $ "Time: " ++ show time
  case res of 
       Nothing -> -- if null ds then 
                     iterRun nameMVar relative maybetimeout proof mvarprob doms dim_bsize_list origl (weaksrs,strictsrs) 
                  --  else
                  --   iterRun nameMVar relative maybetimeout proof mvarprob (ds:doms) ((dim,bsize):dim_bsize_list) origl (weaksrs,strictsrs) 
                    
       Just (d,dim,xs) -> do
          let cIntsS = check_interpretation_of_rules 1 d dim strictsrs xs
          let cIntsW = check_interpretation_of_rules (length cIntsS+1) d dim weaksrs xs
          let (strict,nonstrict) = partition (strictrule d) cIntsS
          let newsrs = map (\(_,rule,_) -> rule) nonstrict
          let neww = map (\(_,rule,_) -> rule) cIntsW
          appendStream proof (unlines 
                               ["Prover applied to",
                                (if relative then T.showSRSRel (weaksrs,strictsrs) else T.showSRS strictsrs),
                                "Found " ++ domainName d ++ " matrix interpretation\n" ++ printAllInterpret d xs
                               , printCheckL d  cIntsS
                               ,if relative then (printCheckLRel d $ cIntsW) else ""
                               ,"strictly decreasing rules:\n"
                               ,printSetOfRules strict
                               ,"weakly decreasing rules:\n"
                               ,printSetOfRules nonstrict
                               ,if relative then (printSetOfRulesRel cIntsW) else ""
                               ,"Remaining problem:"
                               ,if relative then  T.showSRSRel (neww,newsrs) else (T.showSRS newsrs)])
          takeMVar mvarprob
          putMVar mvarprob (neww,newsrs)
          if null strict  then 
             do 
               x <- takeMVar proof
               error ("stopped\n " ++ x )
            else do
                  iterRun nameMVar relative maybetimeout proof mvarprob doms origl origl (neww,newsrs)
          
-- ------------------------------------------------------
-- runP gets a dimension, a width for bitvectors and an SRS 
-- it calls yices to search for trace-decreasing matrix interpretations
-- 
-- it uses temporary files:
--  .tdmi_yices_input is used as input file for yices
--  .tdmi_yices_output is used as output file for yices
-- it deletes both of these files
--

runP nameMVar relative maybetimeout dom dim bwidth (weaksrs,strictsrs) =
  let   
     srs = strictsrs ++ weaksrs
     bitup = maximum $ map length $ concat $ [[l,r] | (l,r) <- srs]
     (conds,syms) = makeConditions relative dom dim (weaksrs,strictsrs)
     smt = fixBitWidth dom ((bwidth+dim-1)*bitup) (conds,syms)  bwidth
  in do
      freshName <- do 
                    ((n:ns),used) <- takeMVar nameMVar
                    putMVar nameMVar (ns,n:used)
                    return n
      let inputname = freshName ++ "_input"
      let outputname = freshName ++ "_output"
      let scriptname = freshName ++ "_script.sh"
      mkCallYicesSMT2 inputname outputname scriptname maybetimeout  
          
      writeFile inputname $ smt 
      u <- runCommand $ "./" ++ scriptname
      waitForProcess u
      b <- doesFileExist outputname -- ".tdmi_yices_output"
      if b then do
        x <- strictReadFile outputname -- ".tdmi_yices_output"
        k <- removeFileSafe outputname -- .tdmi_yices_output"
        k <- removeFileSafe inputname
        k <- removeFileSafe scriptname -- .tdmi_yices_output"
        if "sat" `isPrefixOf` x then return $ Just  (dom,dim,parseYicesOutput x)
                                else return Nothing
       else do
        k <- removeFileSafe outputname -- .tdmi_yices_output"
        k <- removeFileSafe inputname
        k <- removeFileSafe scriptname -- .tdmi_yices_output"
       
        return Nothing 
--  where syscall = sc 
          
          
appendStream :: MVar String -> String -> IO ()
appendStream mvar str =
  do
    buff <- takeMVar mvar
    putMVar mvar (buff++str ++ "\n")
    
prependStream :: MVar String -> String -> IO ()    
prependStream mvar str =
  do
    buff <- takeMVar mvar
    putMVar mvar (str ++ "\n" ++buff)
